from __future__ import annotations

from datetime import datetime, timedelta, timezone

import pytest

from app.core.utils.time import utcnow
from app.db.models import RequestLog
from app.db.session import SessionLocal
from app.modules.api_keys.repository import ApiKeysRepository

pytestmark = pytest.mark.integration


async def _create_api_key(async_client, *, name: str) -> str:
    response = await async_client.post("/api/api-keys/", json={"name": name})
    assert response.status_code == 200
    return response.json()["id"]


async def _insert_request_logs(*rows: RequestLog) -> None:
    async with SessionLocal() as session:
        session.add_all(rows)
        await session.commit()


def _parse_utc(value: str) -> datetime:
    return datetime.fromisoformat(value.replace("Z", "+00:00")).astimezone(timezone.utc)


def _hour_bucket(value: datetime) -> datetime:
    if value.tzinfo is None:
        value = value.replace(tzinfo=timezone.utc)
    else:
        value = value.astimezone(timezone.utc)
    return value.replace(minute=0, second=0, microsecond=0)


@pytest.mark.asyncio
@pytest.mark.parametrize("endpoint", ["trends", "usage-7d"])
async def test_api_key_detail_endpoints_return_404_for_missing_key(async_client, endpoint: str):
    response = await async_client.get(f"/api/api-keys/missing-key/{endpoint}")
    assert response.status_code == 404


@pytest.mark.asyncio
async def test_trends_returns_hourly_zero_filled_points_with_bucket_aggregation(
    async_client,
    monkeypatch: pytest.MonkeyPatch,
):
    key_id = await _create_api_key(async_client, name="trend-key")
    now = datetime(2026, 3, 30, 12, 0, 0)
    monkeypatch.setattr("app.modules.api_keys.service.utcnow", lambda: now)

    aggregated_bucket = now.replace(minute=0, second=0, microsecond=0) - timedelta(hours=5)
    separate_bucket = now.replace(minute=0, second=0, microsecond=0) - timedelta(hours=2)
    same_bucket_a = aggregated_bucket + timedelta(minutes=10)
    same_bucket_b = aggregated_bucket + timedelta(minutes=45)
    other_bucket = separate_bucket + timedelta(minutes=15)
    old_row = now - timedelta(days=8)

    await _insert_request_logs(
        RequestLog(
            api_key_id=key_id,
            request_id="req-trend-a",
            requested_at=same_bucket_a,
            model="gpt-5.1",
            status="ok",
            input_tokens=30,
            output_tokens=12,
            cached_input_tokens=6,
            cost_usd=0.125,
        ),
        RequestLog(
            api_key_id=key_id,
            request_id="req-trend-b",
            requested_at=same_bucket_b,
            model="gpt-5.1",
            status="ok",
            input_tokens=7,
            output_tokens=3,
            cached_input_tokens=2,
            cost_usd=0.075,
        ),
        RequestLog(
            api_key_id=key_id,
            request_id="req-trend-c",
            requested_at=other_bucket,
            model="gpt-5.1",
            status="ok",
            input_tokens=10,
            output_tokens=5,
            cached_input_tokens=1,
            cost_usd=0.05,
        ),
        RequestLog(
            api_key_id=key_id,
            request_id="req-trend-d",
            requested_at=other_bucket + timedelta(minutes=10),
            model="gpt-5.1",
            status="ok",
            input_tokens=4,
            output_tokens=None,
            reasoning_tokens=6,
            cached_input_tokens=0,
            cost_usd=0.02,
        ),
        RequestLog(
            api_key_id=key_id,
            request_id="req-trend-old",
            requested_at=old_row,
            model="gpt-5.1",
            status="ok",
            input_tokens=999,
            output_tokens=999,
            cached_input_tokens=999,
            cost_usd=9.99,
        ),
    )

    response = await async_client.get(f"/api/api-keys/{key_id}/trends")
    assert response.status_code == 200

    payload = response.json()
    assert payload["keyId"] == key_id
    assert len(payload["cost"]) == 168
    assert len(payload["tokens"]) == 168

    cost_times = [_parse_utc(point["t"]) for point in payload["cost"]]
    token_times = [_parse_utc(point["t"]) for point in payload["tokens"]]
    assert cost_times == token_times
    assert cost_times == sorted(cost_times)
    assert cost_times[1] - cost_times[0] == timedelta(hours=1)

    cost_by_bucket = {_parse_utc(point["t"]): point["v"] for point in payload["cost"]}
    tokens_by_bucket = {_parse_utc(point["t"]): point["v"] for point in payload["tokens"]}

    assert tokens_by_bucket[_hour_bucket(same_bucket_a)] == pytest.approx(52.0)
    assert cost_by_bucket[_hour_bucket(same_bucket_a)] == pytest.approx(0.2)
    assert tokens_by_bucket[_hour_bucket(other_bucket)] == pytest.approx(25.0)
    assert cost_by_bucket[_hour_bucket(other_bucket)] == pytest.approx(0.07)

    assert sum(point["v"] for point in payload["tokens"]) == pytest.approx(77.0)
    assert sum(point["v"] for point in payload["cost"]) == pytest.approx(0.27)
    assert sum(1 for point in payload["tokens"] if point["v"] == 0) > 0


@pytest.mark.asyncio
async def test_trends_include_oldest_partial_hour_for_rolling_7d_window(
    async_client,
    monkeypatch: pytest.MonkeyPatch,
):
    key_id = await _create_api_key(async_client, name="trend-partial-window-key")
    now = datetime(2026, 3, 30, 10, 37, 0)
    since = now - timedelta(days=7)
    monkeypatch.setattr("app.modules.api_keys.service.utcnow", lambda: now)

    oldest_included = since + timedelta(minutes=5)
    oldest_excluded = since - timedelta(minutes=1)
    newest_included = now.replace(minute=10, second=0, microsecond=0)
    newest_excluded = now

    await _insert_request_logs(
        RequestLog(
            api_key_id=key_id,
            request_id="req-trend-partial-old-excluded",
            requested_at=oldest_excluded,
            model="gpt-5.1",
            status="ok",
            input_tokens=50,
            output_tokens=5,
            cached_input_tokens=0,
            cost_usd=0.5,
        ),
        RequestLog(
            api_key_id=key_id,
            request_id="req-trend-partial-old-included",
            requested_at=oldest_included,
            model="gpt-5.1",
            status="ok",
            input_tokens=10,
            output_tokens=2,
            cached_input_tokens=0,
            cost_usd=0.1,
        ),
        RequestLog(
            api_key_id=key_id,
            request_id="req-trend-partial-new-included",
            requested_at=newest_included,
            model="gpt-5.1",
            status="ok",
            input_tokens=4,
            output_tokens=None,
            reasoning_tokens=3,
            cached_input_tokens=0,
            cost_usd=0.05,
        ),
        RequestLog(
            api_key_id=key_id,
            request_id="req-trend-partial-new-excluded",
            requested_at=newest_excluded,
            model="gpt-5.1",
            status="ok",
            input_tokens=100,
            output_tokens=20,
            cached_input_tokens=0,
            cost_usd=1.0,
        ),
    )

    response = await async_client.get(f"/api/api-keys/{key_id}/trends")
    assert response.status_code == 200

    payload = response.json()
    assert len(payload["cost"]) == 169
    assert len(payload["tokens"]) == 169

    cost_times = [_parse_utc(point["t"]) for point in payload["cost"]]
    cost_by_bucket = {_parse_utc(point["t"]): point["v"] for point in payload["cost"]}
    tokens_by_bucket = {_parse_utc(point["t"]): point["v"] for point in payload["tokens"]}

    assert cost_times[0] == _hour_bucket(since)
    assert cost_times[-1] == _hour_bucket(now)
    assert tokens_by_bucket[_hour_bucket(oldest_included)] == pytest.approx(12.0)
    assert cost_by_bucket[_hour_bucket(oldest_included)] == pytest.approx(0.1)
    assert tokens_by_bucket[_hour_bucket(newest_included)] == pytest.approx(7.0)
    assert cost_by_bucket[_hour_bucket(newest_included)] == pytest.approx(0.05)
    assert sum(point["v"] for point in payload["tokens"]) == pytest.approx(19.0)
    assert sum(point["v"] for point in payload["cost"]) == pytest.approx(0.15)


@pytest.mark.asyncio
async def test_trends_repository_bounds_results_to_the_requested_window(async_client):
    key_id = await _create_api_key(async_client, name="trend-window-key")
    window_start = datetime(2026, 3, 23, 10, 37, 0)
    window_end = datetime(2026, 3, 30, 10, 37, 0)
    included_oldest = window_start + timedelta(minutes=5)
    included_newest = window_end - timedelta(minutes=30)
    excluded_old = window_start - timedelta(seconds=1)
    excluded_future = window_end

    await _insert_request_logs(
        RequestLog(
            api_key_id=key_id,
            request_id="req-trend-bounded-old",
            requested_at=excluded_old,
            model="gpt-5.1",
            status="ok",
            input_tokens=99,
            output_tokens=1,
            cached_input_tokens=0,
            cost_usd=0.99,
        ),
        RequestLog(
            api_key_id=key_id,
            request_id="req-trend-bounded-a",
            requested_at=included_oldest,
            model="gpt-5.1",
            status="ok",
            input_tokens=10,
            output_tokens=2,
            cached_input_tokens=0,
            cost_usd=0.1,
        ),
        RequestLog(
            api_key_id=key_id,
            request_id="req-trend-bounded-b",
            requested_at=included_newest,
            model="gpt-5.1",
            status="ok",
            input_tokens=4,
            output_tokens=3,
            cached_input_tokens=0,
            cost_usd=0.05,
        ),
        RequestLog(
            api_key_id=key_id,
            request_id="req-trend-bounded-future",
            requested_at=excluded_future,
            model="gpt-5.1",
            status="ok",
            input_tokens=500,
            output_tokens=100,
            cached_input_tokens=0,
            cost_usd=5.0,
        ),
    )

    async with SessionLocal() as session:
        repo = ApiKeysRepository(session)
        buckets = await repo.trends_by_key(key_id, window_start, window_end, 3600)

    assert [bucket.bucket_epoch for bucket in buckets] == [
        int(_hour_bucket(included_oldest).timestamp()),
        int(_hour_bucket(included_newest).timestamp()),
    ]
    assert [bucket.total_tokens for bucket in buckets] == [12, 7]
    assert [bucket.total_cost_usd for bucket in buckets] == [pytest.approx(0.1), pytest.approx(0.05)]


@pytest.mark.asyncio
async def test_usage_7d_sums_only_recent_request_logs(async_client):
    key_id = await _create_api_key(async_client, name="usage-key")
    now = utcnow()

    await _insert_request_logs(
        RequestLog(
            api_key_id=key_id,
            request_id="req-usage-a",
            requested_at=now - timedelta(days=1, hours=2),
            model="gpt-5.1",
            status="ok",
            input_tokens=120,
            output_tokens=35,
            cached_input_tokens=20,
            cost_usd=0.42,
        ),
        RequestLog(
            api_key_id=key_id,
            request_id="req-usage-b",
            requested_at=now - timedelta(hours=8),
            model="gpt-5.1",
            status="error",
            input_tokens=8,
            output_tokens=2,
            cached_input_tokens=0,
            cost_usd=0.08,
        ),
        RequestLog(
            api_key_id=key_id,
            request_id="req-usage-c",
            requested_at=now - timedelta(hours=3),
            model="gpt-5.1",
            status="ok",
            input_tokens=2,
            output_tokens=None,
            reasoning_tokens=7,
            cached_input_tokens=1,
            cost_usd=0.03,
        ),
        RequestLog(
            api_key_id=key_id,
            request_id="req-usage-old",
            requested_at=now - timedelta(days=7, minutes=1),
            model="gpt-5.1",
            status="ok",
            input_tokens=500,
            output_tokens=100,
            cached_input_tokens=50,
            cost_usd=5.0,
        ),
    )

    response = await async_client.get(f"/api/api-keys/{key_id}/usage-7d")
    assert response.status_code == 200

    payload = response.json()
    assert payload == {
        "keyId": key_id,
        "totalTokens": 174,
        "totalCostUsd": 0.53,
        "totalRequests": 3,
        "cachedInputTokens": 21,
    }


@pytest.mark.asyncio
async def test_usage_7d_clamps_cached_input_tokens_to_total_input(async_client):
    key_id = await _create_api_key(async_client, name="usage-key-clamped-cache")
    now = utcnow()

    await _insert_request_logs(
        RequestLog(
            api_key_id=key_id,
            request_id="req-usage-clamp-a",
            requested_at=now - timedelta(hours=4),
            model="gpt-5.1",
            status="ok",
            input_tokens=10,
            output_tokens=3,
            cached_input_tokens=999,
            cost_usd=0.12,
        ),
        RequestLog(
            api_key_id=key_id,
            request_id="req-usage-clamp-b",
            requested_at=now - timedelta(hours=1),
            model="gpt-5.1",
            status="ok",
            input_tokens=4,
            output_tokens=None,
            reasoning_tokens=2,
            cached_input_tokens=8,
            cost_usd=0.03,
        ),
    )

    response = await async_client.get(f"/api/api-keys/{key_id}/usage-7d")
    assert response.status_code == 200

    payload = response.json()
    assert payload == {
        "keyId": key_id,
        "totalTokens": 19,
        "totalCostUsd": 0.15,
        "totalRequests": 2,
        "cachedInputTokens": 14,
    }



from __future__ import annotations

import asyncio
import base64
import json

import pytest

import app.modules.oauth.service as oauth_module
from app.core.auth import generate_unique_account_id
from app.core.clients.oauth import DeviceCode, OAuthTokens
from app.core.crypto import TokenEncryptor
from app.core.utils.time import utcnow
from app.db.models import Account, AccountStatus
from app.db.session import SessionLocal
from app.modules.accounts.repository import AccountsRepository

pytestmark = pytest.mark.integration


def _encode_jwt(payload: dict) -> str:
    raw = json.dumps(payload, separators=(",", ":")).encode("utf-8")
    body = base64.urlsafe_b64encode(raw).rstrip(b"=").decode("ascii")
    return f"header.{body}.sig"


@pytest.mark.asyncio
async def test_device_oauth_flow_creates_account(async_client, monkeypatch):
    await oauth_module._OAUTH_STORE.reset()

    email = "device@example.com"
    raw_account_id = "acc_device"

    async def fake_device_code(**_):
        return DeviceCode(
            verification_url="https://auth.openai.com/codex/device",
            user_code="ABCD-EFGH",
            device_auth_id="dev_123",
            interval_seconds=1,
            expires_in_seconds=30,
        )

    async def fake_exchange_device_token(**_):
        payload = {
            "email": email,
            "chatgpt_account_id": raw_account_id,
            "https://api.openai.com/auth": {"chatgpt_plan_type": "plus"},
        }
        return OAuthTokens(
            access_token="access-token",
            refresh_token="refresh-token",
            id_token=_encode_jwt(payload),
        )

    async def fake_sleep(_: float) -> None:
        return None

    monkeypatch.setattr(oauth_module, "request_device_code", fake_device_code)
    monkeypatch.setattr(oauth_module, "exchange_device_token", fake_exchange_device_token)
    monkeypatch.setattr(oauth_module, "_async_sleep", fake_sleep)

    start = await async_client.post("/api/oauth/start", json={"forceMethod": "device"})
    assert start.status_code == 200
    assert start.json()["method"] == "device"

    complete = await async_client.post("/api/oauth/complete", json={})
    assert complete.status_code == 200
    assert complete.json()["status"] == "pending"

    await asyncio.sleep(0)

    payload = None
    for _ in range(20):
        status = await async_client.get("/api/oauth/status")
        assert status.status_code == 200
        payload = status.json()
        if payload["status"] == "success":
            break
        await asyncio.sleep(0.05)
    assert payload and payload["status"] == "success"

    expected_account_id = generate_unique_account_id(raw_account_id, email)
    accounts = await async_client.get("/api/accounts")
    assert accounts.status_code == 200
    data = accounts.json()["accounts"]
    assert any(account["accountId"] == expected_account_id for account in data)


@pytest.mark.asyncio
async def test_device_oauth_flow_keeps_separate_accounts_when_import_without_overwrite_enabled(
    async_client,
    monkeypatch,
):
    await oauth_module._OAUTH_STORE.reset()

    settings = await async_client.put(
        "/api/settings",
        json={
            "stickyThreadsEnabled": False,
            "preferEarlierResetAccounts": False,
            "importWithoutOverwrite": True,
            "totpRequiredOnLogin": False,
        },
    )
    assert settings.status_code == 200
    assert settings.json()["importWithoutOverwrite"] is True

    email = "device-separate@example.com"
    raw_account_id = "acc_device_separate"

    async def fake_device_code(**_):
        return DeviceCode(
            verification_url="https://auth.openai.com/codex/device",
            user_code="ABCD-EFGH",
            device_auth_id="dev_sep",
            interval_seconds=1,
            expires_in_seconds=30,
        )

    call_count = {"value": 0}

    async def fake_exchange_device_token(**_):
        call_count["value"] += 1
        plan_type = "plus" if call_count["value"] == 1 else "team"
        payload = {
            "email": email,
            "chatgpt_account_id": raw_account_id,
            "https://api.openai.com/auth": {"chatgpt_plan_type": plan_type},
        }
        return OAuthTokens(
            access_token=f"access-token-{call_count['value']}",
            refresh_token=f"refresh-token-{call_count['value']}",
            id_token=_encode_jwt(payload),
        )

    async def fake_sleep(_: float) -> None:
        return None

    monkeypatch.setattr(oauth_module, "request_device_code", fake_device_code)
    monkeypatch.setattr(oauth_module, "exchange_device_token", fake_exchange_device_token)
    monkeypatch.setattr(oauth_module, "_async_sleep", fake_sleep)

    async def _run_device_flow_once() -> None:
        start = await async_client.post("/api/oauth/start", json={"forceMethod": "device"})
        assert start.status_code == 200
        assert start.json()["method"] == "device"

        complete = await async_client.post("/api/oauth/complete", json={})
        assert complete.status_code == 200
        assert complete.json()["status"] == "pending"

        await asyncio.sleep(0)

        payload = None
        for _ in range(20):
            status = await async_client.get("/api/oauth/status")
            assert status.status_code == 200
            payload = status.json()
            if payload["status"] == "success":
                break
            await asyncio.sleep(0.05)
        assert payload and payload["status"] == "success"

    await _run_device_flow_once()
    await _run_device_flow_once()

    accounts = await async_client.get("/api/accounts")
    assert accounts.status_code == 200
    data = [account for account in accounts.json()["accounts"] if account["email"] == email]
    assert len(data) == 2
    ids = {account["accountId"] for account in data}
    base_id = generate_unique_account_id(raw_account_id, email)
    assert base_id in ids
    assert any(account_id.startswith(f"{base_id}__copy") for account_id in ids if account_id != base_id)


@pytest.mark.asyncio
async def test_device_oauth_flow_reports_error_when_duplicate_email_is_ambiguous_in_overwrite_mode(
    async_client,
    monkeypatch,
):
    await oauth_module._OAUTH_STORE.reset()

    enable_separate = await async_client.put(
        "/api/settings",
        json={
            "stickyThreadsEnabled": False,
            "preferEarlierResetAccounts": False,
            "importWithoutOverwrite": True,
            "totpRequiredOnLogin": False,
        },
    )
    assert enable_separate.status_code == 200
    assert enable_separate.json()["importWithoutOverwrite"] is True

    email = "oauth-conflict@example.com"
    raw_account_id = "acc_oauth_conflict_base"

    async def fake_device_code(**_):
        return DeviceCode(
            verification_url="https://auth.openai.com/codex/device",
            user_code="ABCD-EFGH",
            device_auth_id="dev_conflict",
            interval_seconds=1,
            expires_in_seconds=30,
        )

    call_count = {"value": 0}

    async def fake_exchange_device_token(**_):
        call_count["value"] += 1
        if call_count["value"] <= 2:
            account_id = raw_account_id
            plan_type = "plus" if call_count["value"] == 1 else "team"
        else:
            account_id = "acc_oauth_conflict_new"
            plan_type = "pro"
        payload = {
            "email": email,
            "chatgpt_account_id": account_id,
            "https://api.openai.com/auth": {"chatgpt_plan_type": plan_type},
        }
        return OAuthTokens(
            access_token=f"access-token-{call_count['value']}",
            refresh_token=f"refresh-token-{call_count['value']}",
            id_token=_encode_jwt(payload),
        )

    async def fake_sleep(_: float) -> None:
        return None

    monkeypatch.setattr(oauth_module, "request_device_code", fake_device_code)
    monkeypatch.setattr(oauth_module, "exchange_device_token", fake_exchange_device_token)
    monkeypatch.setattr(oauth_module, "_async_sleep", fake_sleep)

    async def _run_device_flow_once() -> dict[str, str | None]:
        start = await async_client.post("/api/oauth/start", json={"forceMethod": "device"})
        assert start.status_code == 200
        assert start.json()["method"] == "device"

        complete = await async_client.post("/api/oauth/complete", json={})
        assert complete.status_code == 200
        assert complete.json()["status"] == "pending"

        await asyncio.sleep(0)

        payload: dict[str, str | None] | None = None
        for _ in range(20):
            status = await async_client.get("/api/oauth/status")
            assert status.status_code == 200
            payload = status.json()
            if payload["status"] in {"success", "error"}:
                break
            await asyncio.sleep(0.05)
        assert payload is not None
        return payload

    assert (await _run_device_flow_once())["status"] == "success"
    assert (await _run_device_flow_once())["status"] == "success"

    enable_overwrite = await async_client.put(
        "/api/settings",
        json={
            "stickyThreadsEnabled": False,
            "preferEarlierResetAccounts": False,
            "importWithoutOverwrite": False,
            "totpRequiredOnLogin": False,
        },
    )
    assert enable_overwrite.status_code == 200
    assert enable_overwrite.json()["importWithoutOverwrite"] is False

    result = await _run_device_flow_once()
    assert result["status"] == "error"
    assert result["errorMessage"] is not None
    assert "multiple matching accounts exist" in str(result["errorMessage"]).lower()


@pytest.mark.asyncio
async def test_oauth_start_with_existing_account_marks_success(async_client):
    await oauth_module._OAUTH_STORE.reset()

    encryptor = TokenEncryptor()
    account = Account(
        id="acc_existing",
        email="existing@example.com",
        plan_type="plus",
        access_token_encrypted=encryptor.encrypt("access"),
        refresh_token_encrypted=encryptor.encrypt("refresh"),
        id_token_encrypted=encryptor.encrypt("id"),
        last_refresh=utcnow(),
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )
    async with SessionLocal() as session:
        repo = AccountsRepository(session)
        await repo.upsert(account)

    start = await async_client.post("/api/oauth/start", json={})
    assert start.status_code == 200
    assert start.json()["method"] == "browser"

    status = await async_client.get("/api/oauth/status")
    assert status.status_code == 200
    assert status.json()["status"] == "success"


@pytest.mark.asyncio
async def test_oauth_start_falls_back_to_device_on_os_error(async_client, monkeypatch):
    await oauth_module._OAUTH_STORE.reset()

    async def fake_browser_flow(self):
        raise OSError("no port")

    async def fake_device_code(**_):
        return DeviceCode(
            verification_url="https://auth.openai.com/codex/device",
            user_code="ABCD-EFGH",
            device_auth_id="dev_fallback",
            interval_seconds=1,
            expires_in_seconds=30,
        )

    monkeypatch.setattr(oauth_module.OauthService, "_start_browser_flow", fake_browser_flow)
    monkeypatch.setattr(oauth_module, "request_device_code", fake_device_code)

    start = await async_client.post("/api/oauth/start", json={})
    assert start.status_code == 200
    payload = start.json()
    assert payload["method"] == "device"
    assert payload["deviceAuthId"] == "dev_fallback"


@pytest.mark.asyncio
async def test_manual_callback_returns_success_and_creates_account(async_client, monkeypatch):
    await oauth_module._OAUTH_STORE.reset()

    async def fake_callback_server_start(self) -> None:
        return None

    email = "manual@example.com"
    raw_account_id = "acc_manual"

    async def fake_exchange_authorization_code(**_):
        payload = {
            "email": email,
            "chatgpt_account_id": raw_account_id,
            "https://api.openai.com/auth": {"chatgpt_plan_type": "plus"},
        }
        return OAuthTokens(
            access_token="manual-access-token",
            refresh_token="manual-refresh-token",
            id_token=_encode_jwt(payload),
        )

    monkeypatch.setattr(oauth_module.OAuthCallbackServer, "start", fake_callback_server_start)
    monkeypatch.setattr(oauth_module, "exchange_authorization_code", fake_exchange_authorization_code)

    start = await async_client.post("/api/oauth/start", json={"forceMethod": "browser"})
    assert start.status_code == 200
    assert start.json()["method"] == "browser"

    async with oauth_module._OAUTH_STORE.lock:
        state_token = oauth_module._OAUTH_STORE.state.state_token

    response = await async_client.post(
        "/api/oauth/manual-callback",
        json={
            "callbackUrl": f"http://localhost:1455/auth/callback?code=manual-code&state={state_token}",
        },
    )
    assert response.status_code == 200
    assert response.json() == {"status": "success", "errorMessage": None}

    status = await async_client.get("/api/oauth/status")
    assert status.status_code == 200
    assert status.json()["status"] == "success"

    expected_account_id = generate_unique_account_id(raw_account_id, email)
    accounts = await async_client.get("/api/accounts")
    assert accounts.status_code == 200
    data = accounts.json()["accounts"]
    assert any(account["accountId"] == expected_account_id for account in data)


@pytest.mark.asyncio
async def test_manual_callback_returns_error_message_for_invalid_state(async_client, monkeypatch):
    await oauth_module._OAUTH_STORE.reset()

    async def fake_callback_server_start(self) -> None:
        return None

    monkeypatch.setattr(oauth_module.OAuthCallbackServer, "start", fake_callback_server_start)

    start = await async_client.post("/api/oauth/start", json={"forceMethod": "browser"})
    assert start.status_code == 200
    assert start.json()["method"] == "browser"

    response = await async_client.post(
        "/api/oauth/manual-callback",
        json={
            "callbackUrl": "http://localhost:1455/auth/callback?code=manual-code&state=wrong",
        },
    )
    assert response.status_code == 200
    assert response.json() == {
        "status": "error",
        "errorMessage": "Invalid OAuth callback: state mismatch or missing code.",
    }

    status = await async_client.get("/api/oauth/status")
    assert status.status_code == 200
    assert status.json() == {
        "status": "error",
        "errorMessage": "Invalid OAuth callback: state mismatch or missing code.",
    }



from __future__ import annotations

import base64
import hashlib
import logging
import secrets
from dataclasses import dataclass
from datetime import datetime, timezone
from urllib.parse import quote, urlencode

import aiohttp
from pydantic import ValidationError

from app.core.auth.models import DeviceCodePayload, OAuthTokenPayload
from app.core.clients.http import get_http_client
from app.core.config.settings import get_settings
from app.core.types import JsonObject
from app.core.utils.request_id import get_request_id

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class DeviceCode:
    verification_url: str
    user_code: str
    device_auth_id: str
    interval_seconds: int
    expires_in_seconds: int


@dataclass(frozen=True)
class OAuthTokens:
    access_token: str
    refresh_token: str
    id_token: str


class OAuthError(Exception):
    def __init__(self, code: str, message: str, status_code: int | None = None) -> None:
        super().__init__(message)
        self.code = code
        self.message = message
        self.status_code = status_code


def pkce_challenge(verifier: str) -> str:
    digest = hashlib.sha256(verifier.encode("utf-8")).digest()
    return base64.urlsafe_b64encode(digest).rstrip(b"=").decode("ascii")


def generate_pkce_pair() -> tuple[str, str]:
    verifier = secrets.token_urlsafe(32)
    return verifier, pkce_challenge(verifier)


def build_authorization_url(
    *,
    state: str,
    code_challenge: str,
    base_url: str | None = None,
    client_id: str | None = None,
    originator: str | None = None,
    redirect_uri: str | None = None,
    scope: str | None = None,
) -> str:
    settings = get_settings()
    auth_base = (base_url or settings.auth_base_url).rstrip("/")
    authorization_scope = scope or _ensure_offline_access(settings.oauth_scope)
    params = {
        "response_type": "code",
        "client_id": client_id or settings.oauth_client_id,
        "redirect_uri": redirect_uri or settings.oauth_redirect_uri,
        "scope": authorization_scope,
        "code_challenge": code_challenge,
        "code_challenge_method": "S256",
        "state": state,
        "id_token_add_organizations": "true",
        "codex_cli_simplified_flow": "true",
        "originator": originator or settings.oauth_originator,
    }
    query = urlencode(params, quote_via=quote)
    return f"{auth_base}/oauth/authorize?{query}"


async def exchange_authorization_code(
    *,
    code: str,
    code_verifier: str,
    redirect_uri: str | None = None,
    base_url: str | None = None,
    client_id: str | None = None,
    timeout_seconds: float | None = None,
    session: aiohttp.ClientSession | None = None,
) -> OAuthTokens:
    settings = get_settings()
    url = f"{(base_url or settings.auth_base_url).rstrip('/')}/oauth/token"
    payload = {
        "grant_type": "authorization_code",
        "client_id": client_id or settings.oauth_client_id,
        "code": code,
        "code_verifier": code_verifier,
        "redirect_uri": redirect_uri or settings.oauth_redirect_uri,
    }
    encoded = urlencode(payload, quote_via=quote)
    timeout = aiohttp.ClientTimeout(total=timeout_seconds or settings.oauth_timeout_seconds)

    client_session = session or get_http_client().session
    headers = {"Content-Type": "application/x-www-form-urlencoded"}
    request_id = get_request_id()
    if request_id:
        headers["x-request-id"] = request_id
    async with client_session.post(
        url,
        data=encoded,
        headers=headers,
        timeout=timeout,
    ) as resp:
        data = await _safe_json(resp)
        try:
            payload = OAuthTokenPayload.model_validate(data)
        except ValidationError as exc:
            logger.warning(
                "OAuth token response invalid request_id=%s",
                get_request_id(),
            )
            raise OAuthError("invalid_response", "OAuth response invalid") from exc
        if resp.status >= 400:
            logger.warning(
                "OAuth token request failed request_id=%s status=%s",
                get_request_id(),
                resp.status,
            )
            raise _oauth_error_from_payload(payload, resp.status)

    return _parse_tokens(payload)


async def request_device_code(
    *,
    base_url: str | None = None,
    client_id: str | None = None,
    timeout_seconds: float | None = None,
    session: aiohttp.ClientSession | None = None,
) -> DeviceCode:
    settings = get_settings()
    auth_base = (base_url or settings.auth_base_url).rstrip("/")
    url = f"{auth_base}/api/accounts/deviceauth/usercode"
    payload = {
        "client_id": client_id or settings.oauth_client_id,
    }
    timeout = aiohttp.ClientTimeout(total=timeout_seconds or settings.oauth_timeout_seconds)

    client_session = session or get_http_client().session
    headers: dict[str, str] = {}
    request_id = get_request_id()
    if request_id:
        headers["x-request-id"] = request_id
    async with client_session.post(url, json=payload, headers=headers, timeout=timeout) as resp:
        data = await _safe_json(resp)
        if resp.status >= 400:
            if resp.status == 404:
                raise OAuthError(
                    "device_auth_unavailable",
                    (
                        "Device code login is not enabled for this Codex server. "
                        "Use the browser login or verify the server URL."
                    ),
                    resp.status,
                )
            logger.warning(
                "Device auth request failed request_id=%s status=%s",
                get_request_id(),
                resp.status,
            )
            raise OAuthError(
                "device_auth_failed",
                f"Device code request failed with status {resp.status}",
                resp.status,
            )
        try:
            payload_data = DeviceCodePayload.model_validate(data)
        except ValidationError as exc:
            logger.warning(
                "Device auth response invalid request_id=%s",
                get_request_id(),
            )
            raise OAuthError("invalid_response", "Device auth response invalid") from exc
    verification_url = f"{auth_base}/codex/device"
    user_code = payload_data.user_code
    device_auth_id = payload_data.device_auth_id
    interval = payload_data.interval if payload_data.interval is not None else 0
    expires_in = payload_data.expires_in or 0
    if expires_in <= 0:
        expires_in = _expires_in_seconds(payload_data.expires_at) or 900

    if not user_code or not device_auth_id:
        raise OAuthError("invalid_response", "Device auth response missing fields")

    return DeviceCode(
        verification_url=verification_url,
        user_code=user_code,
        device_auth_id=device_auth_id,
        interval_seconds=interval,
        expires_in_seconds=expires_in,
    )


async def exchange_device_token(
    *,
    device_auth_id: str,
    user_code: str,
    base_url: str | None = None,
    timeout_seconds: float | None = None,
    session: aiohttp.ClientSession | None = None,
) -> OAuthTokens | None:
    settings = get_settings()
    url = f"{(base_url or settings.auth_base_url).rstrip('/')}/api/accounts/deviceauth/token"
    payload = {"device_auth_id": device_auth_id, "user_code": user_code}
    timeout = aiohttp.ClientTimeout(total=timeout_seconds or settings.oauth_timeout_seconds)

    client_session = session or get_http_client().session
    headers: dict[str, str] = {}
    request_id = get_request_id()
    if request_id:
        headers["x-request-id"] = request_id
    async with client_session.post(url, json=payload, headers=headers, timeout=timeout) as resp:
        data = await _safe_json(resp)
        try:
            payload_data = OAuthTokenPayload.model_validate(data)
        except ValidationError as exc:
            logger.warning(
                "Device token response invalid request_id=%s",
                get_request_id(),
            )
            raise OAuthError("invalid_response", "Device auth response invalid") from exc
        if resp.status in (403, 404):
            return None
        if resp.status >= 400:
            if _is_pending_error(payload_data):
                return None
            logger.warning(
                "Device token request failed request_id=%s status=%s",
                get_request_id(),
                resp.status,
            )
            raise _oauth_error_from_payload(payload_data, resp.status)
        if _is_pending_error(payload_data):
            return None

    if payload_data.authorization_code:
        if not payload_data.code_verifier:
            raise OAuthError("invalid_response", "Device auth response missing code verifier")
        redirect_uri = f"{(base_url or settings.auth_base_url).rstrip('/')}/deviceauth/callback"
        return await exchange_authorization_code(
            code=payload_data.authorization_code,
            code_verifier=payload_data.code_verifier,
            redirect_uri=redirect_uri,
            base_url=base_url,
            client_id=settings.oauth_client_id,
            timeout_seconds=timeout_seconds,
        )

    return _parse_tokens(payload_data)


def _ensure_offline_access(scope: str) -> str:
    if "offline_access" in scope.split():
        return scope
    return f"{scope} offline_access"


def _parse_tokens(payload: OAuthTokenPayload) -> OAuthTokens:
    if not payload.access_token or not payload.refresh_token or not payload.id_token:
        raise OAuthError("invalid_response", "OAuth response missing tokens")
    return OAuthTokens(
        access_token=payload.access_token,
        refresh_token=payload.refresh_token,
        id_token=payload.id_token,
    )


async def _safe_json(resp: aiohttp.ClientResponse) -> JsonObject:
    try:
        data = await resp.json(content_type=None)
    except Exception:
        text = await resp.text()
        return {"error": {"message": text.strip()}}
    return data if isinstance(data, dict) else {"error": {"message": str(data)}}


def _oauth_error_from_payload(payload: OAuthTokenPayload, status_code: int) -> OAuthError:
    code = _extract_error_code(payload) or f"http_{status_code}"
    message = _extract_error_message(payload) or f"OAuth request failed ({status_code})"
    return OAuthError(code, message, status_code)


def _extract_error_code(payload: OAuthTokenPayload) -> str | None:
    error = payload.error
    if isinstance(error, dict):
        code = error.get("code") or error.get("error")
        return code if isinstance(code, str) else None
    if isinstance(error, str):
        return error
    return payload.error_code or payload.code


def _extract_error_message(payload: OAuthTokenPayload) -> str | None:
    error = payload.error
    if isinstance(error, dict):
        message = error.get("message") or error.get("error_description")
        return message if isinstance(message, str) else None
    if isinstance(error, str):
        return payload.error_description or error
    return payload.message


def _is_pending_error(payload: OAuthTokenPayload) -> bool:
    code = _extract_error_code(payload)
    if code in {"authorization_pending", "slow_down"}:
        return True
    status = payload.status
    if status and status.lower() in {"pending", "authorization_pending"}:
        return True
    return False


def _expires_in_seconds(expires_at: str | None) -> int | None:
    if not expires_at:
        return None
    try:
        parsed = datetime.fromisoformat(expires_at.replace("Z", "+00:00"))
    except ValueError:
        return None
    if parsed.tzinfo is None:
        parsed = parsed.replace(tzinfo=timezone.utc)
    now = datetime.now(timezone.utc)
    delta = (parsed - now).total_seconds()
    if delta <= 0:
        return None
    return int(delta)



"""add api_key_usage_reservations and reservation_items tables

Revision ID: 20260218_000000_add_api_key_usage_reservations
Revises: 20260215_000000_add_idx_logs_requested_at
Create Date: 2026-02-18
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

# revision identifiers, used by Alembic.
revision = "20260218_000000_add_api_key_usage_reservations"
down_revision = "20260215_000000_add_idx_logs_requested_at"
branch_labels = None
depends_on = None


def _table_exists(connection: Connection, table_name: str) -> bool:
    inspector = sa.inspect(connection)
    return inspector.has_table(table_name)


def _index_exists(connection: Connection, index_name: str, table_name: str) -> bool:
    inspector = sa.inspect(connection)
    if not inspector.has_table(table_name):
        return False
    indexes = inspector.get_indexes(table_name)
    return any(idx["name"] == index_name for idx in indexes)


def upgrade() -> None:
    bind = op.get_bind()

    if not _table_exists(bind, "api_key_usage_reservations"):
        op.create_table(
            "api_key_usage_reservations",
            sa.Column("id", sa.String(), primary_key=True),
            sa.Column(
                "api_key_id",
                sa.String(),
                sa.ForeignKey("api_keys.id", ondelete="CASCADE"),
                nullable=False,
            ),
            sa.Column("model", sa.String(), nullable=False),
            sa.Column("status", sa.String(), nullable=False),
            sa.Column("input_tokens", sa.BigInteger(), nullable=True),
            sa.Column("output_tokens", sa.BigInteger(), nullable=True),
            sa.Column("cached_input_tokens", sa.BigInteger(), nullable=True),
            sa.Column("cost_microdollars", sa.BigInteger(), nullable=True),
            sa.Column("created_at", sa.DateTime(), server_default=sa.func.now(), nullable=False),
            sa.Column("updated_at", sa.DateTime(), server_default=sa.func.now(), nullable=False),
        )
        if not _index_exists(bind, "idx_api_key_usage_reservations_key_id", "api_key_usage_reservations"):
            op.create_index(
                "idx_api_key_usage_reservations_key_id",
                "api_key_usage_reservations",
                ["api_key_id"],
            )
        if not _index_exists(bind, "idx_api_key_usage_reservations_status", "api_key_usage_reservations"):
            op.create_index(
                "idx_api_key_usage_reservations_status",
                "api_key_usage_reservations",
                ["status"],
            )

    if not _table_exists(bind, "api_key_usage_reservation_items"):
        op.create_table(
            "api_key_usage_reservation_items",
            sa.Column("id", sa.Integer(), primary_key=True, autoincrement=True),
            sa.Column(
                "reservation_id",
                sa.String(),
                sa.ForeignKey("api_key_usage_reservations.id", ondelete="CASCADE"),
                nullable=False,
            ),
            sa.Column(
                "limit_id",
                sa.Integer(),
                sa.ForeignKey("api_key_limits.id", ondelete="CASCADE"),
                nullable=False,
            ),
            sa.Column("limit_type", sa.String(), nullable=False),
            sa.Column("reserved_delta", sa.BigInteger(), nullable=False),
            sa.Column("actual_delta", sa.BigInteger(), nullable=True),
            sa.Column("expected_reset_at", sa.DateTime(), nullable=False),
            sa.Column("created_at", sa.DateTime(), server_default=sa.func.now(), nullable=False),
            sa.Column("updated_at", sa.DateTime(), server_default=sa.func.now(), nullable=False),
            sa.UniqueConstraint("reservation_id", "limit_id", name="uq_reservation_limit"),
        )
        if not _index_exists(bind, "idx_api_key_usage_res_items_reservation_id", "api_key_usage_reservation_items"):
            op.create_index(
                "idx_api_key_usage_res_items_reservation_id",
                "api_key_usage_reservation_items",
                ["reservation_id"],
            )


def downgrade() -> None:
    bind = op.get_bind()

    if _table_exists(bind, "api_key_usage_reservation_items"):
        if _index_exists(bind, "idx_api_key_usage_res_items_reservation_id", "api_key_usage_reservation_items"):
            op.drop_index(
                "idx_api_key_usage_res_items_reservation_id",
                table_name="api_key_usage_reservation_items",
            )
        op.drop_table("api_key_usage_reservation_items")

    if _table_exists(bind, "api_key_usage_reservations"):
        if _index_exists(bind, "idx_api_key_usage_reservations_status", "api_key_usage_reservations"):
            op.drop_index("idx_api_key_usage_reservations_status", table_name="api_key_usage_reservations")
        if _index_exists(bind, "idx_api_key_usage_reservations_key_id", "api_key_usage_reservations"):
            op.drop_index("idx_api_key_usage_reservations_key_id", table_name="api_key_usage_reservations")
        op.drop_table("api_key_usage_reservations")



from __future__ import annotations

from fastapi import APIRouter, Depends, Query

from app.core.auth.dependencies import set_dashboard_error_format, validate_dashboard_session
from app.dependencies import AuditContext, get_audit_context
from app.modules.audit.schemas import AuditLogResponse

router = APIRouter(
    prefix="/api/audit-logs",
    tags=["dashboard"],
    dependencies=[Depends(validate_dashboard_session), Depends(set_dashboard_error_format)],
)


@router.get("", response_model=list[AuditLogResponse])
async def list_audit_logs(
    action: str | None = Query(default=None),
    limit: int = Query(default=50, ge=1, le=200),
    offset: int = Query(default=0, ge=0),
    context: AuditContext = Depends(get_audit_context),
) -> list[AuditLogResponse]:
    rows = await context.service.list_logs(action=action, limit=limit, offset=offset)
    return [
        AuditLogResponse(
            id=row.id,
            timestamp=row.timestamp,
            action=row.action,
            actor_ip=row.actor_ip,
            details=row.details,
            request_id=row.request_id,
        )
        for row in rows
    ]



"""add cache invalidation table

Revision ID: 20260401_000000_add_cache_invalidation
Revises: 20260330_020000_add_bridge_ring_members
Create Date: 2026-04-01
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

# revision identifiers, used by Alembic.
revision = "20260401_000000_add_cache_invalidation"
down_revision = "20260330_020000_add_bridge_ring_members"
branch_labels = None
depends_on = None


def _has_table(connection: Connection, table_name: str) -> bool:
    inspector = sa.inspect(connection)
    return inspector.has_table(table_name)


def upgrade() -> None:
    bind = op.get_bind()
    if _has_table(bind, "cache_invalidation"):
        return

    op.create_table(
        "cache_invalidation",
        sa.Column("namespace", sa.String(50), primary_key=True),
        sa.Column("version", sa.Integer, nullable=False, server_default="0"),
    )

    op.execute(sa.text("INSERT INTO cache_invalidation (namespace, version) VALUES ('api_key', 0), ('firewall', 0)"))


def downgrade() -> None:
    op.drop_table("cache_invalidation")



from __future__ import annotations

from pathlib import Path

import pytest

pytestmark = pytest.mark.integration

_STATIC_DIR = Path(__file__).resolve().parent.parent.parent / "app" / "static"


@pytest.mark.asyncio
async def test_health_endpoint_ok(async_client):
    response = await async_client.get("/health")
    assert response.status_code == 200
    assert response.json() == {"status": "ok"}


@pytest.mark.asyncio
async def test_api_validation_error_returns_dashboard_payload(async_client):
    response = await async_client.get("/api/usage/history?hours=0")
    assert response.status_code == 422
    payload = response.json()
    assert payload["error"]["code"] == "validation_error"
    assert payload["error"]["message"] == "Invalid request payload"


@pytest.mark.asyncio
async def test_api_not_found_returns_dashboard_payload(async_client):
    response = await async_client.get("/api/does-not-exist")
    assert response.status_code == 404
    payload = response.json()
    assert payload["error"]["code"] == "http_404"
    assert payload["error"]["message"] == "Not Found"


@pytest.mark.asyncio
async def test_spa_route_path_returns_index_html(async_client, tmp_path):
    index = _STATIC_DIR / "index.html"
    created = not index.exists()
    if created:
        index.parent.mkdir(parents=True, exist_ok=True)
        index.write_text("<!doctype html><html></html>")
    try:
        response = await async_client.get("/dashboard/settings")
        assert response.status_code == 200
        assert response.headers["content-type"].startswith("text/html")
    finally:
        if created:
            index.unlink(missing_ok=True)


@pytest.mark.asyncio
async def test_missing_static_asset_returns_not_found(async_client):
    response = await async_client.get("/assets/missing.js")
    assert response.status_code == 404
    assert response.json()["detail"] == "Not Found"



from __future__ import annotations

from typing import Literal, TypedDict


class UpstreamError(TypedDict, total=False):
    message: str
    resets_at: int | float
    resets_in_seconds: int | float


FailureClass = Literal["rate_limit", "quota", "retryable_transient", "non_retryable"]
FailurePhase = Literal["connect", "first_event", "mid_stream"]


class ClassifiedFailure(TypedDict):
    failure_class: FailureClass
    phase: FailurePhase
    error_code: str
    error: UpstreamError
    http_status: int | None



from __future__ import annotations

import base64
from typing import cast

import pytest

import app.core.clients.proxy as proxy_module


class FakeContent:
    def __init__(self, chunks: list[bytes]) -> None:
        self._chunks = chunks

    async def iter_chunked(self, size: int):
        for chunk in self._chunks:
            yield chunk


class FakeResponse:
    def __init__(self, status: int, headers: dict[str, str], chunks: list[bytes]) -> None:
        self.status = status
        self.headers = headers
        self.content = FakeContent(chunks)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        return False


class FakeSession:
    def __init__(self, response: FakeResponse) -> None:
        self._response = response
        self.calls: list[dict[str, object]] = []

    def get(
        self,
        url: str,
        timeout=None,
        allow_redirects: bool = False,
        headers: dict[str, str] | None = None,
        server_hostname: str | None = None,
    ):
        self.calls.append(
            {
                "url": url,
                "headers": headers,
                "server_hostname": server_hostname,
                "allow_redirects": allow_redirects,
            }
        )
        return self._response


@pytest.mark.asyncio
async def test_fetch_image_data_url_success(monkeypatch):
    async def resolve_ips(host: str, *, timeout_seconds: float):
        return ["93.184.216.34"]

    monkeypatch.setattr(proxy_module, "_resolve_global_ips", resolve_ips)
    body = b"abc"
    response = FakeResponse(200, {"Content-Type": "image/png"}, [body])
    session = FakeSession(response)

    data_url = await proxy_module._fetch_image_data_url(
        cast(proxy_module.ImageFetchSession, session),
        "https://example.com/a.png",
        1.0,
    )

    expected = "data:image/png;base64," + base64.b64encode(body).decode("ascii")
    assert data_url == expected
    assert session.calls
    first_call = session.calls[0]
    assert first_call["url"] == "https://93.184.216.34/a.png"
    assert first_call["headers"] == {"Host": "example.com"}
    assert first_call["server_hostname"] == "example.com"


@pytest.mark.asyncio
async def test_fetch_image_data_url_failure_status(monkeypatch):
    async def resolve_ips(host: str, *, timeout_seconds: float):
        return ["93.184.216.34"]

    monkeypatch.setattr(proxy_module, "_resolve_global_ips", resolve_ips)
    response = FakeResponse(404, {"Content-Type": "image/png"}, [b"abc"])
    session = FakeSession(response)

    data_url = await proxy_module._fetch_image_data_url(
        cast(proxy_module.ImageFetchSession, session),
        "https://example.com/a.png",
        1.0,
    )

    assert data_url is None


@pytest.mark.asyncio
async def test_fetch_image_data_url_size_limit(monkeypatch):
    monkeypatch.setattr(proxy_module, "_IMAGE_INLINE_MAX_BYTES", 4)

    async def resolve_ips(host: str, *, timeout_seconds: float):
        return ["93.184.216.34"]

    monkeypatch.setattr(proxy_module, "_resolve_global_ips", resolve_ips)
    response = FakeResponse(200, {"Content-Type": "image/png"}, [b"12345"])
    session = FakeSession(response)

    data_url = await proxy_module._fetch_image_data_url(
        cast(proxy_module.ImageFetchSession, session),
        "https://example.com/a.png",
        1.0,
    )

    assert data_url is None


@pytest.mark.parametrize(
    ("url", "expected"),
    [
        ("https://example.com/a.png", True),
        ("http://example.com/a.png", False),
        ("https://127.0.0.1/a.png", False),
        ("https://localhost/a.png", False),
        ("https://100.64.0.1/a.png", False),
        ("https://169.254.169.254/a.png", False),
    ],
)
@pytest.mark.asyncio
async def test_is_safe_image_fetch_url(monkeypatch, url: str, expected: bool):
    async def resolve_ips(host: str, *, timeout_seconds: float):
        return ["93.184.216.34"]

    monkeypatch.setattr(proxy_module, "_resolve_global_ips", resolve_ips)
    assert await proxy_module._is_safe_image_fetch_url(url, connect_timeout=1.0) is expected


@pytest.mark.asyncio
async def test_is_safe_image_fetch_url_blocks_resolved_private_ip(monkeypatch):
    async def resolve_none(host: str, *, timeout_seconds: float):
        return None

    monkeypatch.setattr(proxy_module, "_resolve_global_ips", resolve_none)
    assert await proxy_module._is_safe_image_fetch_url("https://example.com/a.png", connect_timeout=1.0) is False


@pytest.mark.asyncio
async def test_is_safe_image_fetch_url_respects_allowlist(monkeypatch):
    async def resolve_ips(host: str, *, timeout_seconds: float):
        return ["93.184.216.34"]

    monkeypatch.setattr(proxy_module, "_resolve_global_ips", resolve_ips)
    settings = proxy_module.get_settings()
    original = settings.image_inline_allowed_hosts
    original_enabled = settings.image_inline_fetch_enabled
    settings.image_inline_fetch_enabled = True
    settings.image_inline_allowed_hosts = ["allowed.example"]
    try:
        assert await proxy_module._is_safe_image_fetch_url("https://allowed.example/a.png", connect_timeout=1.0)
        assert not await proxy_module._is_safe_image_fetch_url("https://denied.example/a.png", connect_timeout=1.0)
    finally:
        settings.image_inline_fetch_enabled = original_enabled
        settings.image_inline_allowed_hosts = original


@pytest.mark.asyncio
async def test_is_safe_image_fetch_url_blocks_when_feature_disabled(monkeypatch):
    async def resolve_ips(host: str, *, timeout_seconds: float):
        return ["93.184.216.34"]

    monkeypatch.setattr(proxy_module, "_resolve_global_ips", resolve_ips)
    settings = proxy_module.get_settings()
    original_enabled = settings.image_inline_fetch_enabled
    original_hosts = settings.image_inline_allowed_hosts
    settings.image_inline_fetch_enabled = False
    settings.image_inline_allowed_hosts = []
    try:
        assert not await proxy_module._is_safe_image_fetch_url("https://example.com/a.png", connect_timeout=1.0)
    finally:
        settings.image_inline_fetch_enabled = original_enabled
        settings.image_inline_allowed_hosts = original_hosts


@pytest.mark.asyncio
async def test_fetch_image_data_url_uses_fallback_ip_when_first_fails(monkeypatch):
    async def resolve_ips(host: str, *, timeout_seconds: float):
        return ["2001:db8::1", "93.184.216.34"]

    monkeypatch.setattr(proxy_module, "_resolve_global_ips", resolve_ips)

    class FallbackSession:
        def __init__(self) -> None:
            self.calls: list[dict[str, object]] = []
            self._ok_response = FakeResponse(200, {"Content-Type": "image/png"}, [b"ok"])

        def get(
            self,
            url: str,
            timeout=None,
            allow_redirects: bool = False,
            headers: dict[str, str] | None = None,
            server_hostname: str | None = None,
        ):
            self.calls.append({"url": url, "headers": headers, "server_hostname": server_hostname})
            if "[2001:db8::1]" in url:
                raise proxy_module.aiohttp.ClientError("connect failed")
            return self._ok_response

    session = FallbackSession()
    data_url = await proxy_module._fetch_image_data_url(
        cast(proxy_module.ImageFetchSession, session),
        "https://example.com/a.png",
        1.0,
    )

    assert data_url is not None
    assert len(session.calls) == 2
    assert session.calls[0]["url"] == "https://[2001:db8::1]/a.png"
    assert session.calls[1]["url"] == "https://93.184.216.34/a.png"



from __future__ import annotations

import hashlib
import logging

from fastapi import Request, Security
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from starlette.requests import HTTPConnection

from app.core.auth.api_key_cache import get_api_key_cache
from app.core.clients.usage import UsageFetchError, fetch_usage
from app.core.config.settings import get_settings
from app.core.config.settings_cache import get_settings_cache
from app.core.exceptions import DashboardAuthError, ProxyAuthError, ProxyUpstreamError
from app.core.request_locality import is_host_os_request, is_local_request
from app.core.utils.time import utcnow
from app.db.session import get_background_session
from app.modules.accounts.repository import AccountsRepository
from app.modules.api_keys.repository import ApiKeysRepository
from app.modules.api_keys.service import ApiKeyData, ApiKeyInvalidError, ApiKeysService
from app.modules.dashboard_auth.service import DASHBOARD_SESSION_COOKIE, get_dashboard_session_store

logger = logging.getLogger(__name__)

_bearer = HTTPBearer(description="API key (e.g. sk-clb-…)", auto_error=False)


def _insecure_allow_remote_no_auth(settings: object) -> bool:
    return bool(getattr(settings, "insecure_allow_remote_no_auth", False))


def insecure_remote_no_auth_allowed(
    request: HTTPConnection | None,
    *,
    settings: object | None = None,
) -> bool:
    effective_settings = settings or get_settings()
    if not _insecure_allow_remote_no_auth(effective_settings):
        return False
    if request is None:
        return False
    return is_host_os_request(request)


# --- Error format markers ---


def set_openai_error_format(request: Request) -> None:
    request.state.error_format = "openai"


def set_dashboard_error_format(request: Request) -> None:
    request.state.error_format = "dashboard"


# --- Proxy API key auth ---


async def validate_proxy_api_key(
    request: Request,
    credentials: HTTPAuthorizationCredentials | None = Security(_bearer),
) -> ApiKeyData | None:
    authorization = None if credentials is None else f"Bearer {credentials.credentials}"
    return await validate_proxy_api_key_authorization(authorization, request=request)


async def validate_proxy_api_key_authorization(
    authorization: str | None,
    *,
    request: HTTPConnection | None = None,
) -> ApiKeyData | None:
    if insecure_remote_no_auth_allowed(request):
        return None
    settings = await get_settings_cache().get()
    if not settings.api_key_auth_enabled:
        if request is not None and not is_local_request(request):
            raise ProxyAuthError("Proxy authentication must be configured before remote access is allowed")
        return None

    token = _extract_bearer_token(authorization)
    if not token:
        raise ProxyAuthError("Missing API key in Authorization header")

    return await _validate_api_key_token(token)


async def _validate_api_key_token(token: str) -> ApiKeyData:
    """Validate a plain API key token and return the typed key data."""

    token_hash = hashlib.sha256(token.encode("utf-8")).hexdigest()
    cache = get_api_key_cache()
    cached = await cache.get(token_hash)
    if cached is not None:
        if cached.expires_at is not None and cached.expires_at <= utcnow():
            await cache.invalidate(token_hash)
        else:
            return cached

    version_before_read = cache.version
    async with get_background_session() as session:
        service = ApiKeysService(ApiKeysRepository(session))
        try:
            validated = await service.validate_key(token)
            await cache.set(token_hash, validated, if_version=version_before_read)
            return validated
        except ApiKeyInvalidError as exc:
            raise ProxyAuthError(str(exc)) from exc


# --- Self-service usage endpoint auth (always requires valid key) ---


async def validate_usage_api_key(
    request: Request,
    credentials: HTTPAuthorizationCredentials | None = Security(_bearer),
) -> ApiKeyData:
    """Validate API key for self-service usage endpoint.

    Unlike ``validate_proxy_api_key``, this dependency always requires a valid
    Bearer API key, regardless of the global ``api_key_auth_enabled`` setting.
    Raises ProxyAuthError when the key is missing or invalid.
    """
    token = _extract_bearer_token(None if credentials is None else f"Bearer {credentials.credentials}")
    if not token:
        raise ProxyAuthError("Missing API key in Authorization header")

    return await _validate_api_key_token(token)


# --- Dashboard session auth ---


async def validate_dashboard_session(request: Request) -> None:
    dashboard_settings = await get_settings_cache().get()
    password_required = bool(dashboard_settings.password_hash)
    requires_auth = password_required or dashboard_settings.totp_required_on_login
    if (
        insecure_remote_no_auth_allowed(request)
        and not requires_auth
        and not request.url.path.startswith("/api/dashboard-auth")
    ):
        return
    if not requires_auth:
        if not is_local_request(request):
            raise DashboardAuthError(
                "Remote bootstrap is required before dashboard access is allowed",
                code="bootstrap_required",
            )
        return

    if not password_required and dashboard_settings.totp_required_on_login:
        logger.warning(
            "dashboard_auth_migration_inconsistency password_hash is NULL"
            " while totp_required_on_login=true metric=dashboard_auth_migration_inconsistency"
        )

    session_id = request.cookies.get(DASHBOARD_SESSION_COOKIE)
    state = get_dashboard_session_store().get(session_id)
    if state is None:
        raise DashboardAuthError("Authentication is required")
    if password_required and not state.password_verified:
        raise DashboardAuthError("Authentication is required")
    if dashboard_settings.totp_required_on_login and not state.totp_verified:
        raise DashboardAuthError("TOTP verification is required for dashboard access", code="totp_required")


# --- Codex usage caller identity auth ---


async def validate_codex_usage_identity(request: Request) -> None:
    token = _extract_bearer_token(request.headers.get("Authorization"))
    if not token:
        raise ProxyAuthError("Missing ChatGPT token in Authorization header")

    raw_account_id = request.headers.get("chatgpt-account-id")
    account_id = raw_account_id.strip() if raw_account_id else ""
    if not account_id:
        raise ProxyAuthError("Missing chatgpt-account-id header")

    async with get_background_session() as session:
        accounts_repo = AccountsRepository(session)
        is_authorized = await accounts_repo.exists_active_chatgpt_account_id(account_id)
    if not is_authorized:
        raise ProxyAuthError("Unknown or inactive chatgpt-account-id")

    try:
        await fetch_usage(access_token=token, account_id=account_id)
    except UsageFetchError as exc:
        if exc.status_code == 429:
            from app.core.exceptions import ProxyRateLimitError

            raise ProxyRateLimitError(exc.message) from exc
        if exc.status_code in (401, 403):
            raise ProxyAuthError("Invalid ChatGPT token or chatgpt-account-id") from exc
        raise ProxyUpstreamError("Unable to validate ChatGPT credentials at this time") from exc


def _extract_bearer_token(authorization: str | None) -> str | None:
    if authorization is None:
        return None
    prefix = "bearer "
    value = authorization.strip()
    if not value.lower().startswith(prefix):
        return None
    token = value[len(prefix) :].strip()
    if not token:
        return None
    return token



from __future__ import annotations

from dataclasses import dataclass

from app.core import usage as usage_core
from app.core.usage.types import UsageWindowRow
from app.db.models import Account
from app.modules.dashboard.schemas import (
    DashboardOverviewSummary,
    DashboardOverviewTimeframe,
    DashboardOverviewTimeframeKey,
    DashboardUsageCost,
    DashboardUsageMetrics,
)
from app.modules.usage.builders import (
    ActivityCostSummary,
    ActivityMetricsSummary,
    build_usage_window_summary_model,
)


@dataclass(frozen=True)
class DashboardOverviewTimeframeConfig:
    key: DashboardOverviewTimeframeKey
    window_minutes: int
    bucket_seconds: int
    bucket_count: int


_OVERVIEW_TIMEFRAME_CONFIGS: dict[DashboardOverviewTimeframeKey, DashboardOverviewTimeframeConfig] = {
    "1d": DashboardOverviewTimeframeConfig(
        key="1d",
        window_minutes=1440,
        bucket_seconds=3600,
        bucket_count=24,
    ),
    "7d": DashboardOverviewTimeframeConfig(
        key="7d",
        window_minutes=10080,
        bucket_seconds=21600,
        bucket_count=28,
    ),
    "30d": DashboardOverviewTimeframeConfig(
        key="30d",
        window_minutes=43200,
        bucket_seconds=86400,
        bucket_count=30,
    ),
}


def resolve_overview_timeframe(
    key: DashboardOverviewTimeframeKey,
) -> DashboardOverviewTimeframeConfig:
    return _OVERVIEW_TIMEFRAME_CONFIGS[key]


def build_overview_timeframe(
    config: DashboardOverviewTimeframeConfig,
) -> DashboardOverviewTimeframe:
    return DashboardOverviewTimeframe(
        key=config.key,
        windowMinutes=config.window_minutes,
        bucketSeconds=config.bucket_seconds,
        bucketCount=config.bucket_count,
    )


def build_dashboard_overview_summary(
    *,
    accounts: list[Account],
    primary_rows: list[UsageWindowRow],
    secondary_rows: list[UsageWindowRow],
    activity_cost: ActivityCostSummary,
    activity_metrics: ActivityMetricsSummary,
) -> DashboardOverviewSummary:
    account_map = {account.id: account for account in accounts}
    primary_summary = usage_core.summarize_usage_window(primary_rows, account_map, "primary")
    secondary_summary = usage_core.summarize_usage_window(secondary_rows, account_map, "secondary")

    primary_window = build_usage_window_summary_model(usage_core.normalize_usage_window(primary_summary))
    secondary_window = build_usage_window_summary_model(usage_core.normalize_usage_window(secondary_summary))

    return DashboardOverviewSummary(
        primary_window=primary_window,
        secondary_window=secondary_window,
        cost=DashboardUsageCost(
            currency=activity_cost.currency,
            totalUsd=activity_cost.total_usd,
        ),
        metrics=DashboardUsageMetrics(
            requests=activity_metrics.requests,
            tokens=activity_metrics.tokens,
            cached_input_tokens=activity_metrics.cached_input_tokens,
            error_rate=activity_metrics.error_rate,
            error_count=activity_metrics.error_count,
            top_error=activity_metrics.top_error,
        ),
    )



from __future__ import annotations

import base64
import json
from collections.abc import AsyncIterator
from datetime import timezone
from types import SimpleNamespace

import pytest
from fastapi.testclient import TestClient
from sqlalchemy import select
from sqlalchemy.exc import IntegrityError
from starlette.testclient import WebSocketDenialResponse

import app.modules.accounts.service as accounts_service_module
import app.modules.proxy.api as proxy_api_module
import app.modules.proxy.load_balancer as load_balancer_module
import app.modules.proxy.provider_adapters as provider_adapters_module
import app.modules.proxy.service as proxy_service_module
import app.modules.upstream_identities.repository as platform_repository_module
from app.core.clients.openai_platform import (
    OpenAIPlatformError,
    PlatformModelsResponse,
    PlatformResponseResult,
    PlatformStreamResponse,
)
from app.core.crypto import TokenEncryptor
from app.core.openai.model_registry import ReasoningLevel, UpstreamModel, get_model_registry
from app.core.openai.models import OpenAIResponsePayload
from app.core.utils.time import utcnow
from app.db.models import Account, AccountStatus, OpenAIPlatformIdentity, RequestLog
from app.db.session import SessionLocal
from app.modules.usage.repository import UsageRepository

pytestmark = pytest.mark.integration


def _make_upstream_model(slug: str) -> UpstreamModel:
    return UpstreamModel(
        slug=slug,
        display_name=slug,
        description=f"Test model {slug}",
        context_window=128000,
        input_modalities=("text",),
        supported_reasoning_levels=(ReasoningLevel(effort="medium", description="default"),),
        default_reasoning_level="medium",
        supports_reasoning_summaries=False,
        support_verbosity=False,
        default_verbosity=None,
        prefer_websockets=False,
        supports_parallel_tool_calls=True,
        supported_in_api=True,
        minimal_client_version=None,
        priority=0,
        available_in_plans=frozenset({"plus", "pro"}),
        base_instructions="",
        raw={},
    )


async def _populate_platform_model_registry() -> None:
    registry = get_model_registry()
    models = [_make_upstream_model("gpt-5.1"), _make_upstream_model("gpt-5.1-codex")]
    await registry.update({"plus": models, "pro": models})


def _encode_jwt(payload: dict) -> str:
    raw = json.dumps(payload, separators=(",", ":")).encode("utf-8")
    body = base64.urlsafe_b64encode(raw).rstrip(b"=").decode("ascii")
    return f"header.{body}.sig"


def _make_auth_json(account_id: str, email: str) -> dict:
    payload = {
        "email": email,
        "chatgpt_account_id": account_id,
        "https://api.openai.com/auth": {"chatgpt_plan_type": "plus"},
    }
    return {
        "tokens": {
            "idToken": _encode_jwt(payload),
            "accessToken": "access-token",
            "refreshToken": "refresh-token",
            "accountId": account_id,
        },
    }


@pytest.fixture(autouse=True)
def _disable_http_bridge(monkeypatch: pytest.MonkeyPatch) -> None:
    settings = SimpleNamespace(
        http_responses_session_bridge_enabled=False,
        prefer_earlier_reset_accounts=False,
        sticky_reallocation_budget_threshold_pct=95.0,
        sticky_threads_enabled=False,
        openai_cache_affinity_max_age_seconds=300,
        openai_prompt_cache_key_derivation_enabled=True,
        routing_strategy="usage_weighted",
        proxy_request_budget_seconds=75.0,
        compact_request_budget_seconds=75.0,
        transcription_request_budget_seconds=120.0,
        upstream_compact_timeout_seconds=None,
        upstream_stream_transport="auto",
        log_proxy_request_payload=False,
        log_proxy_request_shape=False,
        log_proxy_request_shape_raw_cache_key=False,
        log_proxy_service_tier_trace=False,
        stream_idle_timeout_seconds=300.0,
        drain_primary_threshold_pct=85.0,
        drain_secondary_threshold_pct=90.0,
        platform_fallback_force_enabled=False,
    )

    class _SettingsCache:
        async def get(self):
            return settings

    monkeypatch.setattr(proxy_service_module, "get_settings_cache", lambda: _SettingsCache())
    monkeypatch.setattr(proxy_service_module, "get_settings", lambda: settings)


def _assert_platform_text_input(payload: dict[str, object], expected_text: str) -> None:
    input_value = payload.get("input")
    if input_value == expected_text:
        return
    assert input_value == [
        {
            "role": "user",
            "content": [
                {
                    "type": "input_text",
                    "text": expected_text,
                }
            ],
        }
    ]


async def _import_account(async_client, account_id: str, email: str) -> str:
    files = {"auth_json": ("auth.json", json.dumps(_make_auth_json(account_id, email)), "application/json")}
    response = await async_client.post("/api/accounts/import", files=files)
    assert response.status_code == 200
    return response.json()["accountId"]


async def _seed_usage(account_id: str, *, window: str, used_percent: float, reset_after_seconds: int) -> None:
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    async with SessionLocal() as session:
        result = await session.execute(
            select(Account.id).where((Account.id == account_id) | (Account.chatgpt_account_id == account_id)).limit(1)
        )
        resolved_account_id = result.scalar_one_or_none()
        assert resolved_account_id is not None
        usage_repo = UsageRepository(session)
        await usage_repo.add_entry(
            account_id=resolved_account_id,
            used_percent=used_percent,
            window=window,
            reset_at=now_epoch + reset_after_seconds,
            window_minutes=300 if window == "primary" else 10080,
        )


async def _seed_primary_usage(account_id: str, used_percent: float) -> None:
    await _seed_usage(account_id, window="primary", used_percent=used_percent, reset_after_seconds=3600)


async def _seed_secondary_usage(account_id: str, used_percent: float) -> None:
    await _seed_usage(account_id, window="secondary", used_percent=used_percent, reset_after_seconds=86400)


async def _create_platform_identity(async_client, monkeypatch, *, route_families: list[str] | None = None) -> str:
    async def fake_validate_platform_identity(self, *, api_key, organization=None, project=None):
        del self, api_key, organization, project
        return PlatformModelsResponse(
            payload={
                "object": "list",
                "data": [{"id": "gpt-5.1", "object": "model", "owned_by": "openai"}],
            },
            upstream_request_id="up_req_models_validate",
        )

    monkeypatch.setattr(
        accounts_service_module.OpenAIPlatformProviderAdapter,
        "validate_identity",
        fake_validate_platform_identity,
    )
    response = await async_client.post(
        "/api/accounts/platform",
        json={
            "label": "Platform Key",
            "apiKey": "sk-platform-test",
            "organization": "org_test",
            "project": "proj_test",
            "eligibleRouteFamilies": route_families or ["public_responses_http", "public_models_http"],
        },
    )
    assert response.status_code == 200
    return response.json()["accountId"]


async def _insert_platform_identity_direct(route_families: list[str] | None = None) -> str:
    identity = OpenAIPlatformIdentity(
        id="plat_direct",
        label="Platform Key",
        api_key_encrypted=TokenEncryptor().encrypt("sk-platform-test"),
        organization_id="org_test",
        project_id="proj_test",
        eligible_route_families=",".join(route_families or ["public_responses_http", "public_models_http"]),
        status=AccountStatus.ACTIVE,
        last_validated_at=None,
        last_auth_failure_reason=None,
        deactivation_reason=None,
    )
    async with SessionLocal() as session:
        session.add(identity)
        await session.commit()
    return identity.id


def _platform_identity_payload(route_families: list[str] | None = None) -> dict[str, object]:
    return {
        "label": "Platform Key",
        "apiKey": "sk-platform-test",
        "organization": "org_test",
        "project": "proj_test",
        "eligibleRouteFamilies": route_families or ["public_responses_http", "public_models_http"],
    }


async def _stream_lines(lines: list[str]) -> AsyncIterator[str]:
    for line in lines:
        yield line


async def _latest_request_log() -> RequestLog | None:
    async with SessionLocal() as session:
        result = await session.execute(select(RequestLog).order_by(RequestLog.requested_at.desc()))
        return result.scalars().first()


@pytest.mark.asyncio
async def test_create_and_list_platform_identity(async_client, monkeypatch):
    await _import_account(async_client, "acc_platform_list", "platform-list@example.com")
    identity_id = await _create_platform_identity(async_client, monkeypatch)

    response = await async_client.get("/api/accounts")
    assert response.status_code == 200
    accounts = response.json()["accounts"]
    platform_identity = next(item for item in accounts if item["accountId"] == identity_id)

    assert platform_identity["providerKind"] == "openai_platform"
    assert platform_identity["routingSubjectId"] == identity_id
    assert platform_identity["label"] == "Platform Key"
    assert sorted(platform_identity["eligibleRouteFamilies"]) == [
        "public_models_http",
        "public_responses_http",
    ]
    assert platform_identity["organization"] == "org_test"
    assert platform_identity["project"] == "proj_test"


@pytest.mark.asyncio
async def test_create_platform_identity_requires_existing_chatgpt_account(async_client, monkeypatch):
    async def fake_validate_platform_identity(self, *, api_key, organization=None, project=None):
        del self, api_key, organization, project
        return PlatformModelsResponse(
            payload={"object": "list", "data": [{"id": "gpt-5.1", "object": "model", "owned_by": "openai"}]},
            upstream_request_id="up_req_models_validate",
        )

    monkeypatch.setattr(
        accounts_service_module.OpenAIPlatformProviderAdapter,
        "validate_identity",
        fake_validate_platform_identity,
    )

    response = await async_client.post("/api/accounts/platform", json=_platform_identity_payload())
    assert response.status_code == 400
    assert response.json()["error"]["code"] == "platform_identity_prerequisite_failed"


@pytest.mark.asyncio
async def test_create_platform_identity_rejects_second_platform_key(async_client, monkeypatch):
    await _import_account(async_client, "acc_platform_dupe", "platform-dupe@example.com")
    await _create_platform_identity(async_client, monkeypatch)

    response = await async_client.post(
        "/api/accounts/platform",
        json={
            **_platform_identity_payload(),
            "label": "Second Platform Key",
            "apiKey": "sk-platform-test-2",
        },
    )
    assert response.status_code == 409
    assert response.json()["error"]["code"] == "platform_identity_conflict"


@pytest.mark.asyncio
async def test_create_platform_identity_translates_integrity_conflict(async_client, monkeypatch):
    await _import_account(async_client, "acc_platform_race", "platform-race@example.com")

    async def fake_validate_platform_identity(self, *, api_key, organization=None, project=None):
        del self, api_key, organization, project
        return PlatformModelsResponse(
            payload={"object": "list", "data": [{"id": "gpt-5.1", "object": "model", "owned_by": "openai"}]},
            upstream_request_id="up_req_models_validate",
        )

    async def raise_integrity_error(self, create):
        del self, create
        raise IntegrityError("INSERT INTO openai_platform_identities", {}, Exception("duplicate singleton"))

    monkeypatch.setattr(
        accounts_service_module.OpenAIPlatformProviderAdapter,
        "validate_identity",
        fake_validate_platform_identity,
    )
    monkeypatch.setattr(
        platform_repository_module.OpenAIPlatformIdentitiesRepository,
        "create_identity",
        raise_integrity_error,
    )

    response = await async_client.post("/api/accounts/platform", json=_platform_identity_payload())
    assert response.status_code == 409
    assert response.json()["error"]["code"] == "platform_identity_conflict"


@pytest.mark.asyncio
async def test_v1_models_keeps_chatgpt_primary_when_usage_is_healthy(async_client, monkeypatch):
    account_id = await _import_account(async_client, "acc_models_primary", "models-primary@example.com")
    await _seed_primary_usage(account_id, 10.0)
    await _create_platform_identity(async_client, monkeypatch, route_families=["public_models_http"])

    async def fail_fetch_platform_models(*, base_url, api_key, organization=None, project=None):
        del base_url, api_key, organization, project
        raise AssertionError("healthy ChatGPT pool must stay primary")

    monkeypatch.setattr(provider_adapters_module, "fetch_platform_models", fail_fetch_platform_models)

    response = await async_client.get("/v1/models")
    assert response.status_code == 200
    assert response.json()["object"] == "list"


@pytest.mark.asyncio
async def test_v1_models_falls_back_to_platform_when_primary_usage_is_depleted(async_client, monkeypatch):
    await _populate_platform_model_registry()
    account_id = await _import_account(async_client, "acc_models_fallback", "models-fallback@example.com")
    await _seed_primary_usage(account_id, 95.0)
    identity_id = await _create_platform_identity(async_client, monkeypatch, route_families=["public_models_http"])

    async def fake_fetch_platform_models(*, base_url, api_key, organization=None, project=None):
        del base_url, api_key, organization, project
        return PlatformModelsResponse(
            payload={
                "object": "list",
                "data": [
                    {"id": "gpt-5.1", "object": "model", "owned_by": "openai"},
                    {"id": "gpt-5.1-codex", "object": "model", "owned_by": "openai"},
                ],
            },
            upstream_request_id="up_req_models_1",
        )

    monkeypatch.setattr(provider_adapters_module, "fetch_platform_models", fake_fetch_platform_models)

    response = await async_client.get("/v1/models")
    assert response.status_code == 200
    payload = response.json()
    assert payload["object"] == "list"
    assert [item["id"] for item in payload["data"]] == ["gpt-5.1", "gpt-5.1-codex"]

    log = await _latest_request_log()
    assert log is not None
    assert log.provider_kind == "openai_platform"
    assert log.routing_subject_id == identity_id
    assert log.route_class == "openai_public_http"
    assert log.upstream_request_id == "up_req_models_1"
    assert log.status == "success"


@pytest.mark.asyncio
async def test_v1_models_falls_back_to_platform_when_secondary_usage_is_depleted(async_client, monkeypatch):
    await _populate_platform_model_registry()
    account_id = await _import_account(async_client, "acc_models_secondary_fallback", "models-secondary@example.com")
    await _seed_primary_usage(account_id, 10.0)
    await _seed_secondary_usage(account_id, 95.0)
    identity_id = await _create_platform_identity(async_client, monkeypatch, route_families=["public_models_http"])

    async def fake_fetch_platform_models(*, base_url, api_key, organization=None, project=None):
        del base_url, api_key, organization, project
        return PlatformModelsResponse(
            payload={
                "object": "list",
                "data": [
                    {"id": "gpt-5.1", "object": "model", "owned_by": "openai"},
                ],
            },
            upstream_request_id="up_req_models_secondary_1",
        )

    monkeypatch.setattr(provider_adapters_module, "fetch_platform_models", fake_fetch_platform_models)

    response = await async_client.get("/v1/models")
    assert response.status_code == 200
    payload = response.json()
    assert payload["object"] == "list"
    assert [item["id"] for item in payload["data"]] == ["gpt-5.1"]

    log = await _latest_request_log()
    assert log is not None
    assert log.provider_kind == "openai_platform"
    assert log.routing_subject_id == identity_id
    assert log.upstream_request_id == "up_req_models_secondary_1"
    assert log.status == "success"


@pytest.mark.asyncio
async def test_v1_models_keeps_chatgpt_primary_when_any_candidate_remains_healthy(async_client, monkeypatch):
    await _populate_platform_model_registry()
    drained_account_id = await _import_account(async_client, "acc_models_drained", "models-drained@example.com")
    healthy_account_id = await _import_account(async_client, "acc_models_healthy", "models-healthy@example.com")
    await _seed_primary_usage(drained_account_id, 95.0)
    await _seed_secondary_usage(drained_account_id, 95.0)
    await _seed_primary_usage(healthy_account_id, 10.0)
    await _seed_secondary_usage(healthy_account_id, 10.0)
    await _create_platform_identity(async_client, monkeypatch, route_families=["public_models_http"])

    async def fail_fetch_platform_models(*, base_url, api_key, organization=None, project=None):
        del base_url, api_key, organization, project
        raise AssertionError("healthy ChatGPT candidates must keep the platform fallback idle")

    monkeypatch.setattr(provider_adapters_module, "fetch_platform_models", fail_fetch_platform_models)

    response = await async_client.get("/v1/models")
    assert response.status_code == 200
    assert response.json()["object"] == "list"


@pytest.mark.asyncio
async def test_v1_responses_keeps_chatgpt_primary_when_usage_is_healthy(async_client, monkeypatch):
    raw_account_id = "acc_resp_primary"
    expected_account_id = await _import_account(async_client, raw_account_id, "resp-primary@example.com")
    await _seed_primary_usage(expected_account_id, 10.0)
    await _create_platform_identity(async_client, monkeypatch, route_families=["public_responses_http"])

    async def fail_create_platform_response(*, base_url, payload, api_key, organization=None, project=None):
        del base_url, payload, api_key, organization, project
        raise AssertionError("healthy ChatGPT pool must stay primary")

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kw):
        del payload, headers, access_token, base_url, raise_for_status, _kw
        assert account_id == raw_account_id
        yield (
            'data: {"type":"response.completed","response":{"id":"resp_chatgpt_primary","status":"completed",'
            '"usage":{"input_tokens":2,"output_tokens":3,"total_tokens":5}}}\n\n'
        )

    monkeypatch.setattr(provider_adapters_module, "create_platform_response", fail_create_platform_response)
    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    response = await async_client.post("/v1/responses", json={"model": "gpt-5.1", "input": "hi"})
    assert response.status_code == 200
    payload = response.json()
    assert payload["id"] == "resp_chatgpt_primary"
    assert payload["status"] == "completed"

    log = await _latest_request_log()
    assert log is not None
    assert log.provider_kind == "chatgpt_web"
    assert log.account_id == expected_account_id


@pytest.mark.asyncio
async def test_v1_responses_routes_using_enforced_api_key_model(async_client, monkeypatch):
    raw_account_id = "acc_resp_enforced_model"
    expected_account_id = await _import_account(async_client, raw_account_id, "resp-enforced@example.com")
    await _seed_primary_usage(expected_account_id, 10.0)
    await _create_platform_identity(async_client, monkeypatch, route_families=["public_responses_http"])

    enable = await async_client.put(
        "/api/settings",
        json={
            "stickyThreadsEnabled": False,
            "preferEarlierResetAccounts": False,
            "totpRequiredOnLogin": False,
            "apiKeyAuthEnabled": True,
        },
    )
    assert enable.status_code == 200

    created = await async_client.post(
        "/api/api-keys/",
        json={
            "name": "responses-enforced-model",
            "allowedModels": ["gpt-5.1"],
            "enforcedModel": "gpt-5.1",
        },
    )
    assert created.status_code == 200
    key = created.json()["key"]

    async def fail_create_platform_response(*, base_url, payload, api_key, organization=None, project=None):
        del base_url, payload, api_key, organization, project
        raise AssertionError("enforced ChatGPT model should prevent platform fallback or rejection")

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kw):
        del headers, access_token, base_url, raise_for_status, _kw
        assert payload.model == "gpt-5.1"
        assert account_id == raw_account_id
        yield (
            'data: {"type":"response.completed","response":{"id":"resp_chatgpt_enforced","status":"completed",'
            '"usage":{"input_tokens":2,"output_tokens":3,"total_tokens":5}}}\n\n'
        )

    monkeypatch.setattr(provider_adapters_module, "create_platform_response", fail_create_platform_response)
    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    response = await async_client.post(
        "/v1/responses",
        headers={"Authorization": f"Bearer {key}"},
        json={"model": "platform-only-model", "input": "hi"},
    )
    assert response.status_code == 200
    payload = response.json()
    assert payload["id"] == "resp_chatgpt_enforced"
    assert payload["status"] == "completed"

    log = await _latest_request_log()
    assert log is not None
    assert log.provider_kind == "chatgpt_web"
    assert log.account_id == expected_account_id


@pytest.mark.asyncio
async def test_v1_responses_falls_back_to_platform_for_stateless_requests(async_client, monkeypatch):
    account_id = await _import_account(async_client, "acc_resp_fallback", "resp-fallback@example.com")
    await _seed_primary_usage(account_id, 95.0)
    identity_id = await _create_platform_identity(async_client, monkeypatch, route_families=["public_responses_http"])

    async def fake_create_platform_response(*, base_url, payload, api_key, organization=None, project=None):
        del base_url, api_key, organization, project
        assert payload["model"] == "gpt-5.1"
        assert payload["service_tier"] == "priority"
        _assert_platform_text_input(payload, "hi")
        return PlatformResponseResult(
            payload=OpenAIResponsePayload.model_validate(
                {
                    "id": "resp_platform_1",
                    "status": "completed",
                    "service_tier": "default",
                    "usage": {"input_tokens": 2, "output_tokens": 3, "total_tokens": 5},
                }
            ),
            upstream_request_id="up_req_resp_1",
        )

    monkeypatch.setattr(provider_adapters_module, "create_platform_response", fake_create_platform_response)

    response = await async_client.post(
        "/v1/responses",
        json={"model": "gpt-5.1", "input": "hi", "service_tier": "fast"},
    )
    assert response.status_code == 200
    payload = response.json()
    assert payload["id"] == "resp_platform_1"
    assert payload["status"] == "completed"

    log = await _latest_request_log()
    assert log is not None
    assert log.provider_kind == "openai_platform"
    assert log.routing_subject_id == identity_id
    assert log.route_class == "openai_public_http"
    assert log.upstream_request_id == "up_req_resp_1"
    assert log.requested_service_tier == "priority"
    assert log.actual_service_tier == "default"
    assert log.service_tier == "default"


@pytest.mark.asyncio
async def test_v1_responses_keeps_chatgpt_primary_when_any_account_in_pool_is_healthy(async_client, monkeypatch):
    unhealthy_account_id = await _import_account(async_client, "acc_resp_unhealthy", "resp-unhealthy@example.com")
    healthy_account_id = await _import_account(async_client, "acc_resp_healthy", "resp-healthy@example.com")
    await _seed_primary_usage(unhealthy_account_id, 95.0)
    await _seed_primary_usage(healthy_account_id, 10.0)
    await _create_platform_identity(async_client, monkeypatch, route_families=["public_responses_http"])

    async def fail_create_platform_response(*, base_url, payload, api_key, organization=None, project=None):
        del base_url, payload, api_key, organization, project
        raise AssertionError("a healthy ChatGPT account in the pool must keep platform fallback disabled")

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kw):
        del payload, headers, access_token, base_url, raise_for_status, _kw
        assert account_id == "acc_resp_healthy"
        yield (
            'data: {"type":"response.completed","response":{"id":"resp_chatgpt_pool_healthy","status":"completed",'
            '"usage":{"input_tokens":2,"output_tokens":3,"total_tokens":5}}}\n\n'
        )

    monkeypatch.setattr(provider_adapters_module, "create_platform_response", fail_create_platform_response)
    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    response = await async_client.post("/v1/responses", json={"model": "gpt-5.1", "input": "hi"})
    assert response.status_code == 200
    payload = response.json()
    assert payload["id"] == "resp_chatgpt_pool_healthy"
    assert payload["status"] == "completed"

    log = await _latest_request_log()
    assert log is not None
    assert log.provider_kind == "chatgpt_web"
    assert log.account_id == healthy_account_id


@pytest.mark.asyncio
async def test_v1_responses_falls_back_to_platform_when_secondary_usage_is_depleted(async_client, monkeypatch):
    account_id = await _import_account(async_client, "acc_resp_secondary_fallback", "resp-secondary@example.com")
    await _seed_primary_usage(account_id, 10.0)
    await _seed_secondary_usage(account_id, 95.0)
    identity_id = await _create_platform_identity(async_client, monkeypatch, route_families=["public_responses_http"])

    async def fake_create_platform_response(*, base_url, payload, api_key, organization=None, project=None):
        del base_url, api_key, organization, project
        assert payload["model"] == "gpt-5.1"
        _assert_platform_text_input(payload, "hi")
        return PlatformResponseResult(
            payload=OpenAIResponsePayload.model_validate(
                {
                    "id": "resp_platform_secondary",
                    "status": "completed",
                    "usage": {"input_tokens": 2, "output_tokens": 3, "total_tokens": 5},
                }
            ),
            upstream_request_id="up_req_resp_secondary_1",
        )

    monkeypatch.setattr(provider_adapters_module, "create_platform_response", fake_create_platform_response)

    response = await async_client.post("/v1/responses", json={"model": "gpt-5.1", "input": "hi"})
    assert response.status_code == 200
    payload = response.json()
    assert payload["id"] == "resp_platform_secondary"
    assert payload["status"] == "completed"

    log = await _latest_request_log()
    assert log is not None
    assert log.provider_kind == "openai_platform"
    assert log.routing_subject_id == identity_id
    assert log.route_class == "openai_public_http"
    assert log.upstream_request_id == "up_req_resp_secondary_1"


@pytest.mark.asyncio
async def test_updating_platform_route_families_enables_public_responses_fallback(async_client, monkeypatch):
    account_id = await _import_account(async_client, "acc_resp_route_edit", "resp-route-edit@example.com")
    await _seed_primary_usage(account_id, 95.0)
    identity_id = await _create_platform_identity(async_client, monkeypatch, route_families=["public_models_http"])

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kw):
        del payload, headers, access_token, base_url, raise_for_status, _kw
        assert account_id == "acc_resp_route_edit"
        yield (
            'data: {"type":"response.completed","response":{"id":"resp_chatgpt_before_route_edit","status":"completed",'
            '"usage":{"input_tokens":2,"output_tokens":3,"total_tokens":5}}}\n\n'
        )

    async def fail_create_platform_response(*, base_url, payload, api_key, organization=None, project=None):
        del base_url, payload, api_key, organization, project
        raise AssertionError("platform fallback must stay disabled until public_responses_http is enabled")

    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)
    monkeypatch.setattr(provider_adapters_module, "create_platform_response", fail_create_platform_response)

    initial_response = await async_client.post("/v1/responses", json={"model": "gpt-5.1", "input": "hi"})
    assert initial_response.status_code == 200
    assert initial_response.json()["id"] == "resp_chatgpt_before_route_edit"

    update_response = await async_client.patch(
        f"/api/accounts/platform/{identity_id}",
        json={"eligibleRouteFamilies": ["public_models_http", "public_responses_http"]},
    )
    assert update_response.status_code == 200
    assert sorted(update_response.json()["eligibleRouteFamilies"]) == [
        "public_models_http",
        "public_responses_http",
    ]

    async def fake_create_platform_response(*, base_url, payload, api_key, organization=None, project=None):
        del base_url, api_key, organization, project
        assert payload["model"] == "gpt-5.1"
        _assert_platform_text_input(payload, "hi")
        return PlatformResponseResult(
            payload=OpenAIResponsePayload.model_validate(
                {
                    "id": "resp_platform_after_route_edit",
                    "status": "completed",
                    "usage": {"input_tokens": 2, "output_tokens": 3, "total_tokens": 5},
                }
            ),
            upstream_request_id="up_req_resp_route_edit_1",
        )

    async def fail_stream_after_route_edit(
        payload,
        headers,
        access_token,
        account_id,
        base_url=None,
        raise_for_status=False,
        **_kw,
    ):
        del payload, headers, access_token, account_id, base_url, raise_for_status, _kw
        raise AssertionError("platform fallback should take over after enabling public_responses_http")

    monkeypatch.setattr(provider_adapters_module, "create_platform_response", fake_create_platform_response)
    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fail_stream_after_route_edit)

    fallback_response = await async_client.post("/v1/responses", json={"model": "gpt-5.1", "input": "hi"})
    assert fallback_response.status_code == 200
    assert fallback_response.json()["id"] == "resp_platform_after_route_edit"


@pytest.mark.asyncio
async def test_v1_responses_stream_falls_back_to_platform_when_primary_usage_is_depleted(async_client, monkeypatch):
    account_id = await _import_account(async_client, "acc_resp_stream_fallback", "resp-stream-fallback@example.com")
    await _seed_primary_usage(account_id, 95.0)
    identity_id = await _create_platform_identity(async_client, monkeypatch, route_families=["public_responses_http"])

    async def fake_stream_platform_responses(*, base_url, payload, api_key, organization=None, project=None):
        del base_url, api_key, organization, project
        assert payload["model"] == "gpt-5.1"
        _assert_platform_text_input(payload, "hi")
        return PlatformStreamResponse(
            event_stream=_stream_lines(
                [
                    'data: {"type":"response.created","response":{"id":"resp_platform_stream"}}\n\n',
                    'data: {"type":"response.completed","response":{"id":"resp_platform_stream","status":"completed",'
                    '"usage":{"input_tokens":2,"output_tokens":3,"total_tokens":5}}}\n\n',
                ]
            ),
            upstream_request_id="up_req_resp_stream_1",
        )

    monkeypatch.setattr(provider_adapters_module, "stream_platform_responses", fake_stream_platform_responses)

    async with async_client.stream(
        "POST",
        "/v1/responses",
        json={"model": "gpt-5.1", "input": "hi", "stream": True},
    ) as response:
        assert response.status_code == 200
        lines = [line async for line in response.aiter_lines() if line]

    events = [json.loads(line[6:]) for line in lines if line.startswith("data: ")]
    event_types = [event.get("type") for event in events]
    assert event_types == ["response.created", "response.completed"]

    log = await _latest_request_log()
    assert log is not None
    assert log.provider_kind == "openai_platform"
    assert log.routing_subject_id == identity_id
    assert log.route_class == "openai_public_http"
    assert log.upstream_request_id == "up_req_resp_stream_1"


@pytest.mark.asyncio
async def test_v1_responses_stream_returns_502_when_platform_stream_fails_before_first_event(async_client, monkeypatch):
    account_id = await _import_account(async_client, "acc_resp_stream_bootstrap", "resp-stream-bootstrap@example.com")
    await _seed_primary_usage(account_id, 95.0)
    identity_id = await _create_platform_identity(async_client, monkeypatch, route_families=["public_responses_http"])

    async def failing_stream():
        raise RuntimeError("broken stream bootstrap")
        yield "unreachable"

    async def fake_stream_platform_responses(*, base_url, payload, api_key, organization=None, project=None):
        del base_url, payload, api_key, organization, project
        return PlatformStreamResponse(
            event_stream=failing_stream(),
            upstream_request_id="up_req_resp_stream_bootstrap",
        )

    monkeypatch.setattr(provider_adapters_module, "stream_platform_responses", fake_stream_platform_responses)

    response = await async_client.post(
        "/v1/responses",
        json={"model": "gpt-5.1", "input": "hi", "stream": True},
    )
    assert response.status_code == 502
    payload = response.json()
    assert payload["error"]["code"] == "upstream_unavailable"

    log = await _latest_request_log()
    assert log is not None
    assert log.provider_kind == "openai_platform"
    assert log.routing_subject_id == identity_id
    assert log.error_code == "upstream_unavailable"
    assert log.rejection_reason == "platform_stream_start_failed"


@pytest.mark.asyncio
async def test_v1_chat_completions_stays_on_chatgpt_even_when_platform_fallback_exists(async_client, monkeypatch):
    raw_account_id = "acc_chat_completion_primary"
    expected_account_id = await _import_account(async_client, raw_account_id, "chat-completion-primary@example.com")
    await _seed_primary_usage(expected_account_id, 95.0)
    await _seed_secondary_usage(expected_account_id, 95.0)
    await _create_platform_identity(async_client, monkeypatch, route_families=["public_responses_http"])

    async def fail_stream_platform_responses(*, base_url, payload, api_key, organization=None, project=None):
        del base_url, payload, api_key, organization, project
        raise AssertionError("chat completions must stay on the ChatGPT path in phase 1")

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kw):
        del payload, headers, access_token, base_url, raise_for_status, _kw
        assert account_id == raw_account_id
        yield 'data: {"type":"response.output_text.delta","delta":"hi"}\n\n'
        yield (
            'data: {"type":"response.completed","response":{"id":"resp_chat_completion_primary",'
            '"usage":{"input_tokens":2,"output_tokens":3,"total_tokens":5}}}\n\n'
        )

    monkeypatch.setattr(provider_adapters_module, "stream_platform_responses", fail_stream_platform_responses)
    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    response = await async_client.post(
        "/v1/chat/completions",
        json={"model": "gpt-5.1", "messages": [{"role": "user", "content": "hi"}]},
    )
    assert response.status_code == 200
    payload = response.json()
    assert payload["id"] == "resp_chat_completion_primary"
    assert payload["object"] == "chat.completion"

    log = await _latest_request_log()
    assert log is not None
    assert log.provider_kind == "chatgpt_web"
    assert log.account_id == expected_account_id


@pytest.mark.asyncio
async def test_backend_codex_models_uses_platform_when_backend_http_usage_is_drained(async_client, monkeypatch):
    await _populate_platform_model_registry()
    account_id = await _import_account(async_client, "acc_backend_models_primary", "backend-models-primary@example.com")
    await _seed_primary_usage(account_id, 95.0)
    await _seed_secondary_usage(account_id, 95.0)
    identity_id = await _create_platform_identity(async_client, monkeypatch, route_families=["backend_codex_http"])

    async def fake_fetch_platform_models(*, base_url, api_key, organization=None, project=None):
        del base_url, api_key, organization, project
        return PlatformModelsResponse(
            payload={
                "object": "list",
                "data": [
                    {"id": "gpt-5.1", "object": "model", "owned_by": "openai"},
                    {"id": "gpt-5.1-codex", "object": "model", "owned_by": "openai"},
                ],
            },
            upstream_request_id="up_req_backend_codex_models",
        )

    async def fail_build_codex_models_response(*args, **kwargs):
        del args, kwargs
        raise AssertionError("backend codex models should use the Platform fallback path when backend HTTP is enabled")

    monkeypatch.setattr(provider_adapters_module, "fetch_platform_models", fake_fetch_platform_models)
    monkeypatch.setattr(proxy_api_module, "_build_codex_models_response", fail_build_codex_models_response)

    response = await async_client.get("/backend-api/codex/models")
    assert response.status_code == 200
    payload = response.json()
    assert [entry["slug"] for entry in payload["models"]] == ["gpt-5.1", "gpt-5.1-codex"]

    log = await _latest_request_log()
    assert log is not None
    assert log.provider_kind == "openai_platform"
    assert log.routing_subject_id == identity_id
    assert log.route_class == "chatgpt_private"


@pytest.mark.asyncio
async def test_backend_codex_websocket_stays_on_chatgpt_even_when_platform_fallback_exists(
    async_client,
    app_instance,
    monkeypatch,
):
    await _import_account(async_client, "acc_backend_ws_primary", "backend-ws-primary@example.com")
    await _create_platform_identity(async_client, monkeypatch, route_families=["backend_codex_http"])

    async def allow_firewall(_websocket):
        return None

    async def allow_proxy_auth(_authorization, request=None):
        del _authorization, request
        return None

    async def fake_proxy_responses_websocket(self, websocket, forwarded_headers, **kwargs):
        del self, forwarded_headers, kwargs
        await websocket.send_text(json.dumps({"type": "response.created"}))
        await websocket.close(code=1000)

    monkeypatch.setattr(proxy_api_module, "_websocket_firewall_denial_response", allow_firewall)
    monkeypatch.setattr(proxy_api_module, "validate_proxy_api_key_authorization", allow_proxy_auth)
    monkeypatch.setattr(
        proxy_service_module.ProxyService,
        "proxy_responses_websocket",
        fake_proxy_responses_websocket,
    )

    with TestClient(app_instance) as client:
        with client.websocket_connect("/backend-api/codex/responses") as websocket:
            payload = json.loads(websocket.receive_text())

    assert payload["type"] == "response.created"


@pytest.mark.asyncio
async def test_v1_chat_completions_rejects_platform_only_operation(async_client, monkeypatch):
    await _insert_platform_identity_direct(route_families=["public_responses_http"])

    def fail_stream_responses(self, *args, **kwargs):
        del self, args, kwargs
        raise AssertionError("platform-only chat completions rejection must not start ChatGPT transport")

    monkeypatch.setattr(proxy_service_module.ProxyService, "stream_responses", fail_stream_responses)

    response = await async_client.post(
        "/v1/chat/completions",
        json={"model": "gpt-5.1", "messages": [{"role": "user", "content": "hi"}]},
    )
    assert response.status_code == 400
    assert response.json()["error"]["code"] == "provider_feature_unsupported"


@pytest.mark.asyncio
async def test_v1_models_rejects_platform_only_public_operation(async_client, monkeypatch):
    await _insert_platform_identity_direct(route_families=["public_models_http"])

    async def fail_build_models_response(*args, **kwargs):
        del args, kwargs
        raise AssertionError("platform-only public-route rejection must not fall through to ChatGPT models")

    monkeypatch.setattr(proxy_api_module, "_build_models_response", fail_build_models_response)

    response = await async_client.get("/v1/models")
    assert response.status_code == 400
    payload = response.json()
    assert payload["error"]["code"] == "provider_fallback_requires_chatgpt"


@pytest.mark.asyncio
async def test_v1_models_platform_fallback_respects_api_key_allowed_models(async_client, monkeypatch):
    await _populate_platform_model_registry()
    account_id = await _import_account(async_client, "acc_models_restricted", "models-restricted@example.com")
    await _seed_primary_usage(account_id, 95.0)
    await _create_platform_identity(async_client, monkeypatch, route_families=["public_models_http"])

    enable = await async_client.put(
        "/api/settings",
        json={
            "stickyThreadsEnabled": False,
            "preferEarlierResetAccounts": False,
            "totpRequiredOnLogin": False,
            "apiKeyAuthEnabled": True,
        },
    )
    assert enable.status_code == 200

    created = await async_client.post(
        "/api/api-keys/",
        json={
            "name": "platform-models-restricted",
            "allowedModels": ["gpt-5.1"],
        },
    )
    assert created.status_code == 200
    key = created.json()["key"]

    async def fake_fetch_platform_models(*, base_url, api_key, organization=None, project=None):
        del base_url, api_key, organization, project
        return PlatformModelsResponse(
            payload={
                "object": "list",
                "data": [
                    {"id": "gpt-5.1", "object": "model", "owned_by": "openai"},
                    {"id": "gpt-5.1-codex", "object": "model", "owned_by": "openai"},
                ],
            },
            upstream_request_id="up_req_models_restricted",
        )

    monkeypatch.setattr(provider_adapters_module, "fetch_platform_models", fake_fetch_platform_models)

    response = await async_client.get("/v1/models", headers={"Authorization": f"Bearer {key}"})
    assert response.status_code == 200
    assert [item["id"] for item in response.json()["data"]] == ["gpt-5.1"]


@pytest.mark.asyncio
async def test_v1_models_platform_auth_failure_updates_identity_state(async_client, monkeypatch):
    await _populate_platform_model_registry()
    account_id = await _import_account(async_client, "acc_models_auth_fail", "models-auth-fail@example.com")
    await _seed_primary_usage(account_id, 95.0)
    identity_id = await _create_platform_identity(async_client, monkeypatch, route_families=["public_models_http"])

    async def fail_fetch_platform_models(*, base_url, api_key, organization=None, project=None):
        del base_url, api_key, organization, project
        raise OpenAIPlatformError(
            401,
            {
                "error": {
                    "code": "invalid_api_key",
                    "message": "Invalid API key",
                }
            },
        )

    monkeypatch.setattr(provider_adapters_module, "fetch_platform_models", fail_fetch_platform_models)

    response = await async_client.get("/v1/models")
    assert response.status_code == 401

    accounts_response = await async_client.get("/api/accounts")
    assert accounts_response.status_code == 200
    platform_account = next(
        account for account in accounts_response.json()["accounts"] if account["accountId"] == identity_id
    )
    assert platform_account["status"] == "deactivated"
    assert platform_account["lastAuthFailureReason"] == "Invalid API key"


@pytest.mark.asyncio
async def test_v1_responses_rejects_platform_only_public_operation(async_client, monkeypatch):
    await _insert_platform_identity_direct(route_families=["public_responses_http"])

    async def fail_collect_responses(*args, **kwargs):
        del args, kwargs
        raise AssertionError("platform-only public-route rejection must not fall through to ChatGPT responses")

    monkeypatch.setattr(proxy_api_module, "_collect_responses", fail_collect_responses)

    response = await async_client.post("/v1/responses", json={"model": "gpt-5.1", "input": "hi"})
    assert response.status_code == 400
    payload = response.json()
    assert payload["error"]["code"] == "provider_fallback_requires_chatgpt"


@pytest.mark.asyncio
async def test_v1_responses_rejects_previous_response_id_when_only_platform(async_client, monkeypatch):
    await _insert_platform_identity_direct(route_families=["public_responses_http"])

    async def fail_create_platform_response(self, *args, **kwargs):
        del self, args, kwargs
        raise AssertionError("continuity rejection must not start upstream transport")

    monkeypatch.setattr(proxy_service_module.ProxyService, "create_platform_response", fail_create_platform_response)

    response = await async_client.post(
        "/v1/responses",
        json={
            "model": "gpt-5.1",
            "input": "hi",
            "previous_response_id": "resp_prev_1",
        },
    )
    assert response.status_code == 400
    payload = response.json()
    assert payload["error"]["code"] == "provider_continuity_unsupported"
    assert payload["error"]["param"] == "previous_response_id"


@pytest.mark.asyncio
async def test_v1_responses_rejects_conversation_when_only_platform_before_upstream_transport(
    async_client,
    monkeypatch,
):
    await _insert_platform_identity_direct(route_families=["public_responses_http"])

    async def fail_create_platform_response(self, *args, **kwargs):
        del self, args, kwargs
        raise AssertionError("continuity rejection must not start upstream transport")

    monkeypatch.setattr(proxy_service_module.ProxyService, "create_platform_response", fail_create_platform_response)

    response = await async_client.post(
        "/v1/responses",
        json={
            "model": "gpt-5.1",
            "input": "hi",
            "conversation": "conv_1",
        },
    )
    assert response.status_code == 400
    payload = response.json()
    assert payload["error"]["code"] == "provider_continuity_unsupported"
    assert payload["error"]["param"] == "conversation"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "header_name",
    ["session_id", "x-codex-session-id", "x-codex-conversation-id", "x-codex-turn-state"],
)
async def test_v1_responses_rejects_continuity_headers_when_only_platform_before_upstream_transport(
    async_client,
    monkeypatch,
    header_name: str,
):
    await _insert_platform_identity_direct(route_families=["public_responses_http"])

    async def fail_create_platform_response(self, *args, **kwargs):
        del self, args, kwargs
        raise AssertionError("continuity rejection must not start upstream transport")

    monkeypatch.setattr(proxy_service_module.ProxyService, "create_platform_response", fail_create_platform_response)

    response = await async_client.post(
        "/v1/responses",
        headers={header_name: "sid_1"},
        json={"model": "gpt-5.1", "input": "hi"},
    )
    assert response.status_code == 400
    payload = response.json()
    assert payload["error"]["code"] == "provider_continuity_unsupported"
    assert payload["error"]["param"] == header_name


@pytest.mark.asyncio
async def test_platform_only_rejects_compact_and_backend_codex_unsupported_routes(async_client, monkeypatch):
    await _insert_platform_identity_direct(route_families=["public_responses_http", "backend_codex_http"])

    async def fail_compact_responses(self, *args, **kwargs):
        del self, args, kwargs
        raise AssertionError("rejected compact path must not start upstream transport")

    def fail_stream_http_responses(self, *args, **kwargs):
        del self, args, kwargs
        raise AssertionError("rejected backend codex path must not start upstream transport")

    monkeypatch.setattr(proxy_service_module.ProxyService, "compact_responses", fail_compact_responses)
    monkeypatch.setattr(proxy_service_module.ProxyService, "stream_http_responses", fail_stream_http_responses)

    compact_response = await async_client.post(
        "/v1/responses/compact",
        json={"model": "gpt-5.1", "input": "hi"},
    )
    assert compact_response.status_code == 400
    assert compact_response.json()["error"]["code"] == "provider_feature_unsupported"

    backend_compact_response = await async_client.post(
        "/backend-api/codex/responses/compact",
        json={"model": "gpt-5.1", "instructions": "hi", "input": []},
    )
    assert backend_compact_response.status_code == 400
    assert backend_compact_response.json()["error"]["code"] == "provider_feature_unsupported"

    backend_response = await async_client.post(
        "/backend-api/codex/responses",
        json={"model": "gpt-5.1", "instructions": "hi", "input": [], "stream": True},
    )
    assert backend_response.status_code == 400
    assert backend_response.json()["error"]["code"] == "provider_fallback_requires_chatgpt"


@pytest.mark.parametrize(
    ("path", "route_families", "expected_code", "expected_param"),
    [
        ("/v1/responses", ["public_responses_http"], "provider_transport_unsupported", "transport"),
        ("/backend-api/codex/responses", ["backend_codex_http"], "provider_transport_unsupported", "transport"),
    ],
)
@pytest.mark.asyncio
async def test_platform_only_websocket_routes_reject_before_upstream_transport(
    async_client,
    app_instance,
    monkeypatch,
    path: str,
    route_families: list[str],
    expected_code: str,
    expected_param: str | None,
):
    async def allow_firewall(_websocket):
        return None

    async def allow_proxy_auth(_authorization, request=None):
        del _authorization, request
        return None

    async def fail_proxy_responses_websocket(self, *args, **kwargs):
        del self, args, kwargs
        raise AssertionError("rejected websocket path must not start upstream transport")

    await _insert_platform_identity_direct(route_families=route_families)
    monkeypatch.setattr(proxy_api_module, "_websocket_firewall_denial_response", allow_firewall)
    monkeypatch.setattr(proxy_api_module, "validate_proxy_api_key_authorization", allow_proxy_auth)
    monkeypatch.setattr(
        proxy_service_module.ProxyService,
        "proxy_responses_websocket",
        fail_proxy_responses_websocket,
    )

    with TestClient(app_instance) as client:
        with pytest.raises(WebSocketDenialResponse) as excinfo:
            with client.websocket_connect(path):
                pass

    denial = excinfo.value
    assert denial.status_code == 400
    payload = denial.json()
    assert payload["error"]["code"] == expected_code
    if expected_param is None:
        assert "param" not in payload["error"]
    else:
        assert payload["error"]["param"] == expected_param


@pytest.mark.asyncio
async def test_backend_codex_models_require_chatgpt_when_only_platform_is_available(async_client, monkeypatch):
    await _insert_platform_identity_direct(route_families=["backend_codex_http"])

    async def fail_build_codex_models_response(*args, **kwargs):
        del args, kwargs
        raise AssertionError("platform-only backend codex models rejection must not fall through to ChatGPT")

    monkeypatch.setattr(proxy_api_module, "_build_codex_models_response", fail_build_codex_models_response)

    response = await async_client.get("/backend-api/codex/models")
    assert response.status_code == 400
    assert response.json()["error"]["code"] == "provider_fallback_requires_chatgpt"


@pytest.mark.asyncio
async def test_backend_codex_responses_uses_platform_when_backend_http_usage_is_drained(
    async_client,
    monkeypatch,
):
    account_id = await _import_account(async_client, "acc_backend_http_primary", "backend-http-primary@example.com")
    await _seed_primary_usage(account_id, 95.0)
    await _seed_secondary_usage(account_id, 95.0)
    identity_id = await _create_platform_identity(async_client, monkeypatch, route_families=["backend_codex_http"])

    async def fake_stream_platform_responses(*, base_url, payload, api_key, organization=None, project=None):
        del base_url, api_key, organization, project
        assert payload["model"] == "gpt-5.1"
        return PlatformStreamResponse(
            event_stream=_stream_lines(
                [
                    'data: {"type":"response.created"}\n\n',
                    'data: {"type":"response.completed","response":{"id":"resp_backend_http_platform"}}\n\n',
                ]
            ),
            upstream_request_id="up_req_backend_http_stream",
        )

    def fail_stream_http_responses(self, *args, **kwargs):
        del self, args, kwargs
        raise AssertionError(
            "backend codex responses should not stay on the ChatGPT path when backend HTTP fallback is enabled"
        )

    monkeypatch.setattr(provider_adapters_module, "stream_platform_responses", fake_stream_platform_responses)
    monkeypatch.setattr(proxy_service_module.ProxyService, "stream_http_responses", fail_stream_http_responses)

    async with async_client.stream(
        "POST",
        "/backend-api/codex/responses",
        json={
            "model": "gpt-5.1",
            "instructions": "hi",
            "input": [],
            "stream": True,
            "reasoning": {"effort": "high"},
        },
    ) as response:
        assert response.status_code == 200
        lines = [line async for line in response.aiter_lines() if line]

    assert lines == [
        'data: {"type":"response.created"}',
        'data: {"type":"response.completed","response":{"id":"resp_backend_http_platform"}}',
    ]

    log = await _latest_request_log()
    assert log is not None
    assert log.provider_kind == "openai_platform"
    assert log.routing_subject_id == identity_id
    assert log.route_class == "chatgpt_private"
    assert log.reasoning_effort == "high"


@pytest.mark.asyncio
async def test_backend_codex_responses_uses_platform_when_force_fallback_is_enabled(
    async_client,
    monkeypatch,
):
    account_id = await _import_account(async_client, "acc_backend_http_force", "backend-http-force@example.com")
    await _seed_primary_usage(account_id, 10.0)
    await _seed_secondary_usage(account_id, 10.0)
    identity_id = await _create_platform_identity(async_client, monkeypatch, route_families=["backend_codex_http"])

    settings = proxy_service_module.get_settings()
    settings.platform_fallback_force_enabled = True
    monkeypatch.setattr(load_balancer_module, "get_settings", lambda: settings)

    async def fake_stream_platform_responses(*, base_url, payload, api_key, organization=None, project=None):
        del base_url, api_key, organization, project
        assert payload["model"] == "gpt-5.1"
        return PlatformStreamResponse(
            event_stream=_stream_lines(
                [
                    'data: {"type":"response.created"}\n\n',
                    'data: {"type":"response.completed","response":{"id":"resp_backend_http_platform_force"}}\n\n',
                ]
            ),
            upstream_request_id="up_req_backend_http_force_stream",
        )

    def fail_stream_http_responses(self, *args, **kwargs):
        del self, args, kwargs
        raise AssertionError("force fallback should bypass the ChatGPT transport for backend codex responses")

    monkeypatch.setattr(provider_adapters_module, "stream_platform_responses", fake_stream_platform_responses)
    monkeypatch.setattr(proxy_service_module.ProxyService, "stream_http_responses", fail_stream_http_responses)

    async with async_client.stream(
        "POST",
        "/backend-api/codex/responses",
        json={"model": "gpt-5.1", "instructions": "hi", "input": [], "stream": True},
    ) as response:
        assert response.status_code == 200
        lines = [line async for line in response.aiter_lines() if line]

    assert lines == [
        'data: {"type":"response.created"}',
        'data: {"type":"response.completed","response":{"id":"resp_backend_http_platform_force"}}',
    ]

    log = await _latest_request_log()
    assert log is not None
    assert log.provider_kind == "openai_platform"
    assert log.routing_subject_id == identity_id
    assert log.route_class == "chatgpt_private"


@pytest.mark.asyncio
async def test_backend_codex_responses_with_turn_state_stays_on_chatgpt_path(async_client, monkeypatch):
    raw_account_id = "acc_backend_http_turn_state"
    expected_account_id = await _import_account(async_client, raw_account_id, "backend-http-turn-state@example.com")
    await _seed_primary_usage(expected_account_id, 95.0)
    await _seed_secondary_usage(expected_account_id, 95.0)
    await _create_platform_identity(async_client, monkeypatch, route_families=["backend_codex_http"])

    async def fail_stream_platform_responses(*, base_url, payload, api_key, organization=None, project=None):
        del base_url, payload, api_key, organization, project
        raise AssertionError("turn-state continuity should keep backend codex responses on the ChatGPT path")

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kw):
        del payload, headers, access_token, base_url, raise_for_status, _kw
        assert account_id == raw_account_id
        yield 'data: {"type":"response.completed","response":{"id":"resp_chatgpt_turn_state"}}\n\n'

    monkeypatch.setattr(provider_adapters_module, "stream_platform_responses", fail_stream_platform_responses)
    monkeypatch.setattr(provider_adapters_module, "core_stream_responses", fake_stream)

    async with async_client.stream(
        "POST",
        "/backend-api/codex/responses",
        headers={"x-codex-turn-state": "turn_state_chatgpt"},
        json={"model": "gpt-5.1", "instructions": "hi", "input": [], "stream": True},
    ) as response:
        assert response.status_code == 200
        lines = [line async for line in response.aiter_lines() if line]

    assert lines == ['data: {"type":"response.completed","response":{"id":"resp_chatgpt_turn_state"}}']

    log = await _latest_request_log()
    assert log is not None
    assert log.provider_kind == "chatgpt_web"
    assert log.account_id == expected_account_id


@pytest.mark.asyncio
async def test_v1_responses_platform_logs_enforced_api_key_service_tier(async_client, monkeypatch):
    account_id = await _import_account(async_client, "acc_platform_enforced_tier", "platform-enforced-tier@example.com")
    await _seed_primary_usage(account_id, 95.0)
    await _seed_secondary_usage(account_id, 95.0)
    await _create_platform_identity(async_client, monkeypatch, route_families=["public_responses_http"])

    enable = await async_client.put(
        "/api/settings",
        json={
            "stickyThreadsEnabled": False,
            "preferEarlierResetAccounts": False,
            "totpRequiredOnLogin": False,
            "apiKeyAuthEnabled": True,
        },
    )
    assert enable.status_code == 200

    created = await async_client.post(
        "/api/api-keys/",
        json={
            "name": "platform-enforced-service-tier",
            "allowedModels": ["gpt-5.1"],
            "enforcedModel": "gpt-5.1",
            "enforcedServiceTier": "fast",
        },
    )
    assert created.status_code == 200
    key = created.json()["key"]
    assert created.json()["enforcedServiceTier"] == "priority"

    async def fake_create_platform_response(*, base_url, payload, api_key, organization=None, project=None):
        del base_url, api_key, organization, project
        assert payload["model"] == "gpt-5.1"
        assert payload["service_tier"] == "priority"
        _assert_platform_text_input(payload, "hi")
        return PlatformResponseResult(
            payload=OpenAIResponsePayload.model_validate(
                {
                    "id": "resp_platform_enforced_tier",
                    "status": "completed",
                    "usage": {"input_tokens": 2, "output_tokens": 3, "total_tokens": 5},
                }
            ),
            upstream_request_id="up_req_platform_enforced_tier",
        )

    monkeypatch.setattr(provider_adapters_module, "create_platform_response", fake_create_platform_response)

    response = await async_client.post(
        "/v1/responses",
        headers={"Authorization": f"Bearer {key}"},
        json={"model": "gpt-5.1", "input": "hi", "service_tier": "default"},
    )
    assert response.status_code == 200
    assert response.json()["id"] == "resp_platform_enforced_tier"

    log = await _latest_request_log()
    assert log is not None
    assert log.provider_kind == "openai_platform"
    assert log.requested_service_tier == "priority"
    assert log.actual_service_tier is None
    assert log.service_tier == "priority"


@pytest.mark.asyncio
async def test_backend_codex_responses_rejects_continuity_when_only_platform(async_client, monkeypatch):
    await _insert_platform_identity_direct(route_families=["backend_codex_http"])

    async def fail_create_platform_response(self, *args, **kwargs):
        del self, args, kwargs
        raise AssertionError("continuity rejection must not start platform transport")

    monkeypatch.setattr(proxy_service_module.ProxyService, "create_platform_response", fail_create_platform_response)

    response = await async_client.post(
        "/backend-api/codex/responses",
        json={
            "model": "gpt-5.1",
            "instructions": "hi",
            "input": [],
            "previous_response_id": "resp_prev_1",
        },
    )
    assert response.status_code == 400
    payload = response.json()
    assert payload["error"]["code"] == "provider_continuity_unsupported"
    assert payload["error"]["param"] == "previous_response_id"



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import cast as typing_cast

import anyio
from sqlalchemy import Integer, String, and_, cast, func, literal_column, or_, select
from sqlalchemy import exc as sa_exc
from sqlalchemy.ext.asyncio import AsyncSession

from app.core.usage.logs import RequestLogLike, calculated_cost_from_log
from app.core.usage.types import BucketModelAggregate, RequestActivityAggregate
from app.core.utils.request_id import ensure_request_id
from app.core.utils.time import utcnow
from app.db.models import Account, ApiKey, RequestLog
from app.modules.upstream_identities.types import CHATGPT_WEB_PROVIDER_KIND


@dataclass(frozen=True, slots=True)
class _RequestLogFilters:
    conditions: list
    needs_related_search_joins: bool


class RequestLogsRepository:
    def __init__(self, session: AsyncSession) -> None:
        self._session = session

    async def list_since(self, since: datetime) -> list[RequestLog]:
        result = await self._session.execute(select(RequestLog).where(RequestLog.requested_at >= since))
        return list(result.scalars().all())

    async def aggregate_by_bucket(
        self,
        since: datetime,
        bucket_seconds: int = 21600,
    ) -> list[BucketModelAggregate]:
        bind = self._session.get_bind()
        dialect = bind.dialect.name if bind else "sqlite"
        if dialect == "postgresql":
            bucket_expr = func.floor(func.extract("epoch", RequestLog.requested_at) / bucket_seconds) * bucket_seconds
        else:
            # Use explicit integer division for SQLite: CAST(epoch / N AS INTEGER) * N
            epoch_col = cast(func.strftime("%s", RequestLog.requested_at), Integer)
            bucket_expr = cast(epoch_col / bucket_seconds, Integer) * bucket_seconds
        bucket_col = bucket_expr.label("bucket_epoch")

        stmt = (
            select(
                bucket_col,
                RequestLog.model,
                RequestLog.service_tier,
                func.count().label("request_count"),
                func.sum(cast(RequestLog.status != literal_column("'success'"), Integer)).label("error_count"),
                func.coalesce(func.sum(RequestLog.input_tokens), 0).label("input_tokens"),
                func.coalesce(func.sum(RequestLog.output_tokens), 0).label("output_tokens"),
                func.coalesce(func.sum(RequestLog.cached_input_tokens), 0).label("cached_input_tokens"),
                func.coalesce(func.sum(RequestLog.reasoning_tokens), 0).label("reasoning_tokens"),
                func.coalesce(func.sum(RequestLog.cost_usd), 0.0).label("cost_usd"),
            )
            .where(RequestLog.requested_at >= since)
            .group_by(bucket_col, RequestLog.model, RequestLog.service_tier)
            .order_by(bucket_col)
        )
        result = await self._session.execute(stmt)
        return [
            BucketModelAggregate(
                bucket_epoch=int(row.bucket_epoch),
                model=row.model,
                service_tier=row.service_tier,
                request_count=int(row.request_count),
                error_count=int(row.error_count),
                input_tokens=int(row.input_tokens),
                output_tokens=int(row.output_tokens),
                cached_input_tokens=int(row.cached_input_tokens),
                reasoning_tokens=int(row.reasoning_tokens),
                cost_usd=float(row.cost_usd or 0.0),
            )
            for row in result.all()
        ]

    async def aggregate_activity_since(self, since: datetime) -> RequestActivityAggregate:
        stmt = select(
            func.count().label("request_count"),
            func.coalesce(
                func.sum(cast(RequestLog.status != literal_column("'success'"), Integer)),
                0,
            ).label("error_count"),
            func.coalesce(func.sum(RequestLog.input_tokens), 0).label("input_tokens"),
            func.coalesce(func.sum(RequestLog.output_tokens), 0).label("output_tokens"),
            func.coalesce(func.sum(RequestLog.cached_input_tokens), 0).label("cached_input_tokens"),
            func.coalesce(func.sum(RequestLog.cost_usd), 0.0).label("cost_usd"),
        ).where(RequestLog.requested_at >= since)
        result = await self._session.execute(stmt)
        row = result.one()
        return RequestActivityAggregate(
            request_count=int(row.request_count),
            error_count=int(row.error_count),
            input_tokens=int(row.input_tokens),
            output_tokens=int(row.output_tokens),
            cached_input_tokens=int(row.cached_input_tokens),
            cost_usd=float(row.cost_usd or 0.0),
        )

    async def top_error_since(self, since: datetime) -> str | None:
        stmt = (
            select(RequestLog.error_code, func.count(RequestLog.id).label("error_count"))
            .where(
                RequestLog.requested_at >= since,
                RequestLog.status != "success",
                RequestLog.error_code.is_not(None),
            )
            .group_by(RequestLog.error_code)
            .order_by(func.count(RequestLog.id).desc(), RequestLog.error_code.asc())
            .limit(1)
        )
        result = await self._session.execute(stmt)
        row = result.first()
        return str(row[0]) if row and row[0] else None

    async def add_log(
        self,
        account_id: str | None,
        *,
        provider_kind: str | None = None,
        routing_subject_id: str | None = None,
        request_id: str,
        model: str,
        input_tokens: int | None,
        output_tokens: int | None,
        latency_ms: int | None,
        status: str,
        error_code: str | None,
        latency_first_token_ms: int | None = None,
        error_message: str | None = None,
        requested_at: datetime | None = None,
        cached_input_tokens: int | None = None,
        reasoning_tokens: int | None = None,
        reasoning_effort: str | None = None,
        service_tier: str | None = None,
        requested_service_tier: str | None = None,
        actual_service_tier: str | None = None,
        transport: str | None = None,
        api_key_id: str | None = None,
        route_class: str | None = None,
        upstream_request_id: str | None = None,
        rejection_reason: str | None = None,
    ) -> RequestLog:
        resolved_request_id = ensure_request_id(request_id)
        resolved_provider_kind = provider_kind
        if resolved_provider_kind is None and account_id is not None:
            resolved_provider_kind = CHATGPT_WEB_PROVIDER_KIND
        resolved_routing_subject_id = routing_subject_id or account_id
        log = RequestLog(
            account_id=account_id,
            provider_kind=resolved_provider_kind,
            routing_subject_id=resolved_routing_subject_id,
            api_key_id=api_key_id,
            request_id=resolved_request_id,
            model=model,
            transport=transport,
            route_class=route_class,
            upstream_request_id=upstream_request_id,
            rejection_reason=rejection_reason,
            service_tier=service_tier,
            requested_service_tier=requested_service_tier,
            actual_service_tier=actual_service_tier,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            cached_input_tokens=cached_input_tokens,
            reasoning_tokens=reasoning_tokens,
            cost_usd=None,
            reasoning_effort=reasoning_effort,
            latency_ms=latency_ms,
            latency_first_token_ms=latency_first_token_ms,
            status=status,
            error_code=error_code,
            error_message=error_message,
            requested_at=requested_at or utcnow(),
        )
        log.cost_usd = calculated_cost_from_log(typing_cast(RequestLogLike, log))
        self._session.add(log)
        try:
            await self._session.commit()
            await self._session.refresh(log)
            return log
        except sa_exc.ResourceClosedError:
            return log
        except BaseException:
            await _safe_rollback(self._session)
            raise

    async def list_recent(
        self,
        limit: int = 50,
        offset: int = 0,
        search: str | None = None,
        since: datetime | None = None,
        until: datetime | None = None,
        account_ids: list[str] | None = None,
        model_options: list[tuple[str, str | None]] | None = None,
        models: list[str] | None = None,
        reasoning_efforts: list[str] | None = None,
        include_success: bool = True,
        include_error_other: bool = True,
        error_codes_in: list[str] | None = None,
        error_codes_excluding: list[str] | None = None,
    ) -> tuple[list[RequestLog], int]:
        filters = self._build_filters(
            search=search,
            since=since,
            until=until,
            account_ids=account_ids,
            model_options=model_options,
            models=models,
            reasoning_efforts=reasoning_efforts,
            include_success=include_success,
            include_error_other=include_error_other,
            error_codes_in=error_codes_in,
            error_codes_excluding=error_codes_excluding,
        )

        total_col = func.count().over().label("_total")
        stmt = select(RequestLog, total_col).order_by(RequestLog.requested_at.desc(), RequestLog.id.desc())
        stmt = self._apply_related_search_joins(stmt, filters.needs_related_search_joins)
        if filters.conditions:
            stmt = stmt.where(and_(*filters.conditions))
        if offset:
            stmt = stmt.offset(offset)
        if limit:
            stmt = stmt.limit(limit)
        result = await self._session.execute(stmt)
        rows = result.all()
        if not rows:
            return [], await self._count_recent(filters)
        logs = [row[0] for row in rows]
        total = rows[0][1]
        return logs, total

    async def _count_recent(self, filters: _RequestLogFilters) -> int:
        count_stmt = select(func.count(RequestLog.id)).select_from(RequestLog)
        count_stmt = self._apply_related_search_joins(count_stmt, filters.needs_related_search_joins)
        if filters.conditions:
            count_stmt = count_stmt.where(and_(*filters.conditions))
        result = await self._session.execute(count_stmt)
        return int(result.scalar_one())

    async def list_filter_options(
        self,
        since: datetime | None = None,
        until: datetime | None = None,
        account_ids: list[str] | None = None,
        model_options: list[tuple[str, str | None]] | None = None,
        models: list[str] | None = None,
        reasoning_efforts: list[str] | None = None,
    ) -> tuple[list[str], list[tuple[str, str | None]], list[tuple[str, str | None]]]:
        filters = self._build_filters(
            since=since,
            until=until,
            account_ids=account_ids,
            model_options=model_options,
            models=models,
            reasoning_efforts=reasoning_efforts,
            include_success=True,
            include_error_other=True,
            error_codes_in=None,
            error_codes_excluding=None,
        )

        subject_id = _request_log_subject_id_expr()
        account_stmt = select(subject_id).distinct().order_by(subject_id.asc())
        model_stmt = (
            select(RequestLog.model, RequestLog.reasoning_effort)
            .distinct()
            .order_by(RequestLog.model.asc(), RequestLog.reasoning_effort.asc())
        )
        status_stmt = (
            select(RequestLog.status, RequestLog.error_code)
            .distinct()
            .order_by(RequestLog.status.asc(), RequestLog.error_code.asc())
        )
        if filters.conditions:
            clause = and_(*filters.conditions)
            account_stmt = account_stmt.where(clause)
            model_stmt = model_stmt.where(clause)
            status_stmt = status_stmt.where(clause)

        account_rows = await self._session.execute(account_stmt)
        model_rows = await self._session.execute(model_stmt)
        status_rows = await self._session.execute(status_stmt)

        account_ids = [row[0] for row in account_rows.all() if row[0]]
        model_options = [(row[0], row[1]) for row in model_rows.all() if row[0]]
        status_values = [(row[0], row[1]) for row in status_rows.all() if row[0]]
        return account_ids, model_options, status_values

    async def get_api_key_names_by_ids(self, api_key_ids: list[str]) -> dict[str, str]:
        unique_ids = sorted({key_id for key_id in api_key_ids if key_id})
        if not unique_ids:
            return {}
        result = await self._session.execute(select(ApiKey.id, ApiKey.name).where(ApiKey.id.in_(unique_ids)))
        return {key_id: name for key_id, name in result.all() if key_id and name}

    def _build_filters(
        self,
        *,
        search: str | None = None,
        since: datetime | None = None,
        until: datetime | None = None,
        account_ids: list[str] | None = None,
        model_options: list[tuple[str, str | None]] | None = None,
        models: list[str] | None = None,
        reasoning_efforts: list[str] | None = None,
        include_success: bool = True,
        include_error_other: bool = True,
        error_codes_in: list[str] | None = None,
        error_codes_excluding: list[str] | None = None,
    ) -> _RequestLogFilters:
        conditions = []
        if since is not None:
            conditions.append(RequestLog.requested_at >= since)
        if until is not None:
            conditions.append(RequestLog.requested_at <= until)
        if account_ids:
            conditions.append(_request_log_subject_id_expr().in_(account_ids))

        if model_options:
            pair_conditions = []
            for model, effort in model_options:
                base = (model or "").strip()
                if not base:
                    continue
                if effort is None:
                    pair_conditions.append(and_(RequestLog.model == base, RequestLog.reasoning_effort.is_(None)))
                else:
                    pair_conditions.append(and_(RequestLog.model == base, RequestLog.reasoning_effort == effort))
            if pair_conditions:
                conditions.append(or_(*pair_conditions))
        else:
            if models:
                conditions.append(RequestLog.model.in_(models))
            if reasoning_efforts:
                conditions.append(RequestLog.reasoning_effort.in_(reasoning_efforts))

        status_conditions = []
        if include_success:
            status_conditions.append(RequestLog.status == "success")
        if error_codes_in:
            status_conditions.append(and_(RequestLog.status == "error", RequestLog.error_code.in_(error_codes_in)))
        if include_error_other:
            error_clause = [RequestLog.status == "error"]
            if error_codes_excluding:
                error_clause.append(
                    or_(
                        RequestLog.error_code.is_(None),
                        ~RequestLog.error_code.in_(error_codes_excluding),
                    )
                )
            status_conditions.append(and_(*error_clause))
        if status_conditions:
            conditions.append(or_(*status_conditions))
        if search:
            search_pattern = f"%{search}%"
            conditions.append(
                or_(
                    RequestLog.account_id.ilike(search_pattern),
                    RequestLog.routing_subject_id.ilike(search_pattern),
                    Account.email.ilike(search_pattern),
                    RequestLog.request_id.ilike(search_pattern),
                    RequestLog.model.ilike(search_pattern),
                    RequestLog.reasoning_effort.ilike(search_pattern),
                    RequestLog.status.ilike(search_pattern),
                    RequestLog.error_code.ilike(search_pattern),
                    RequestLog.error_message.ilike(search_pattern),
                    RequestLog.api_key_id.ilike(search_pattern),
                    ApiKey.name.ilike(search_pattern),
                    cast(RequestLog.requested_at, String).ilike(search_pattern),
                    cast(RequestLog.input_tokens, String).ilike(search_pattern),
                    cast(RequestLog.output_tokens, String).ilike(search_pattern),
                    cast(RequestLog.cached_input_tokens, String).ilike(search_pattern),
                    cast(RequestLog.reasoning_tokens, String).ilike(search_pattern),
                    cast(RequestLog.latency_ms, String).ilike(search_pattern),
                )
            )
            return _RequestLogFilters(conditions=conditions, needs_related_search_joins=True)
        return _RequestLogFilters(conditions=conditions, needs_related_search_joins=False)

    def _apply_related_search_joins(self, stmt, include_related_search_joins: bool):
        if not include_related_search_joins:
            return stmt
        return stmt.outerjoin(Account, Account.id == RequestLog.account_id).outerjoin(
            ApiKey,
            ApiKey.id == RequestLog.api_key_id,
        )


def _request_log_subject_id_expr():
    return func.coalesce(RequestLog.account_id, RequestLog.routing_subject_id)


async def _safe_rollback(session: AsyncSession) -> None:
    if not session.in_transaction():
        return
    try:
        with anyio.CancelScope(shield=True):
            await session.rollback()
    except BaseException:
        return



from __future__ import annotations

import pytest

from app.core.utils.sse import format_sse_event

pytestmark = pytest.mark.unit


def test_format_sse_event_serializes_payload():
    payload = {"type": "response.completed", "response": {"id": "resp_1"}}
    result = format_sse_event(payload)
    assert result == 'event: response.completed\ndata: {"type":"response.completed","response":{"id":"resp_1"}}\n\n'



from __future__ import annotations

from dataclasses import dataclass

import aiohttp
from aiohttp_retry import RetryClient

from app.core.config.settings import get_settings


@dataclass(slots=True)
class HttpClient:
    session: aiohttp.ClientSession
    websocket_session: aiohttp.ClientSession
    retry_client: RetryClient


_http_client: HttpClient | None = None


async def init_http_client() -> HttpClient:
    global _http_client
    if _http_client is not None:
        return _http_client
    settings = get_settings()
    connector = aiohttp.TCPConnector(
        limit=settings.http_connector_limit,
        limit_per_host=settings.http_connector_limit_per_host,
    )
    session = aiohttp.ClientSession(
        connector=connector,
        timeout=aiohttp.ClientTimeout(total=None),
        trust_env=True,
    )
    try:
        # Match Codex CLI's direct websocket transport by avoiding env proxies unless operators
        # explicitly opt in for websocket handshakes.
        websocket_session = aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(total=None),
            trust_env=settings.upstream_websocket_trust_env,
        )
    except Exception:
        await session.close()
        raise
    retry_client = RetryClient(client_session=session, raise_for_status=False)
    _http_client = HttpClient(
        session=session,
        websocket_session=websocket_session,
        retry_client=retry_client,
    )
    return _http_client


async def close_http_client() -> None:
    global _http_client
    if _http_client is None:
        return
    client = _http_client
    try:
        await client.websocket_session.close()
    finally:
        await client.retry_client.close()
    _http_client = None


def get_http_client() -> HttpClient:
    if _http_client is None:
        raise RuntimeError("HTTP client not initialized")
    return _http_client



from __future__ import annotations

import pytest

from app.core.utils.retry import parse_retry_after

pytestmark = pytest.mark.unit


def test_parse_retry_after_seconds():
    assert parse_retry_after("Try again in 1.2s") == 1.2


def test_parse_retry_after_milliseconds():
    assert parse_retry_after("Try again in 500ms") == 0.5


def test_parse_retry_after_missing():
    assert parse_retry_after("no retry info") is None



from __future__ import annotations

import argparse
import os
import sys
from pathlib import Path
from typing import Sequence

import uvicorn

from app.cli_runtime import (
    DEFAULT_SHUTDOWN_TIMEOUT_SECONDS,
    DEFAULT_STARTUP_TIMEOUT_SECONDS,
    ServeOptions,
    default_log_file,
    default_pid_file,
    load_running_metadata,
    shutdown_background_server,
    start_background_server,
)
from app.core.runtime_logging import build_log_config


def _add_serve_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("--host", default=os.getenv("HOST", "127.0.0.1"))
    parser.add_argument("--port", type=int, default=int(os.getenv("PORT", "2455")))
    parser.add_argument("--ssl-certfile", default=os.getenv("SSL_CERTFILE"))
    parser.add_argument("--ssl-keyfile", default=os.getenv("SSL_KEYFILE"))


def _build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Run the codex-lb-cinamon API server.")
    subparsers = parser.add_subparsers(dest="command", required=True)

    serve_parser = subparsers.add_parser("serve", help="Run the API server in the foreground.")
    _add_serve_arguments(serve_parser)

    start_parser = subparsers.add_parser("start", help="Start the API server in the background.")
    _add_serve_arguments(start_parser)
    start_parser.add_argument("--pid-file", type=Path, default=default_pid_file())
    start_parser.add_argument("--log-file", type=Path, default=default_log_file())
    start_parser.add_argument("--startup-timeout", type=float, default=DEFAULT_STARTUP_TIMEOUT_SECONDS)

    status_parser = subparsers.add_parser("status", help="Show background server status.")
    status_parser.add_argument("--pid-file", type=Path, default=default_pid_file())

    shutdown_parser = subparsers.add_parser("shutdown", help="Stop the tracked background server.")
    shutdown_parser.add_argument("--pid-file", type=Path, default=default_pid_file())
    shutdown_parser.add_argument("--timeout", type=float, default=DEFAULT_SHUTDOWN_TIMEOUT_SECONDS)

    return parser


def _parse_args(argv: Sequence[str] | None = None) -> argparse.Namespace:
    raw_args = list(argv) if argv is not None else sys.argv[1:]
    if raw_args and raw_args[0] in {"-h", "--help"}:
        return _build_parser().parse_args(raw_args)
    if not raw_args or raw_args[0].startswith("-"):
        raw_args = ["serve", *raw_args]
    return _build_parser().parse_args(raw_args)


def _serve_options_from_args(args: argparse.Namespace) -> ServeOptions:
    return ServeOptions(
        host=args.host,
        port=args.port,
        ssl_certfile=args.ssl_certfile,
        ssl_keyfile=args.ssl_keyfile,
    )


def _validate_ssl_flags(options: ServeOptions) -> None:
    if bool(options.ssl_certfile) ^ bool(options.ssl_keyfile):
        raise SystemExit("Both --ssl-certfile and --ssl-keyfile must be provided together.")


def _run_foreground(options: ServeOptions) -> None:
    _validate_ssl_flags(options)
    uvicorn.run(
        "app.main:app",
        host=options.host,
        port=options.port,
        ssl_certfile=options.ssl_certfile,
        ssl_keyfile=options.ssl_keyfile,
        log_config=build_log_config(),
    )


def _run_background_start(args: argparse.Namespace) -> None:
    options = _serve_options_from_args(args)
    _validate_ssl_flags(options)
    try:
        metadata = start_background_server(
            options,
            pid_file=args.pid_file,
            log_file=args.log_file,
            startup_timeout_seconds=args.startup_timeout,
        )
    except RuntimeError as exc:
        raise SystemExit(str(exc)) from exc

    print(f"Started codex-lb-cinamon in background (pid {metadata.pid}, {metadata.host}:{metadata.port})")
    print(f"PID file: {args.pid_file.expanduser()}")
    print(f"Log file: {Path(metadata.log_file).expanduser()}")


def _run_status(args: argparse.Namespace) -> None:
    metadata, stale = load_running_metadata(args.pid_file)
    if metadata is None:
        if stale:
            print(f"No running background server found. Removed stale PID file {args.pid_file.expanduser()}.")
        else:
            print("codex-lb-cinamon background server is not running.")
        raise SystemExit(1)

    print(f"codex-lb-cinamon background server is running (pid {metadata.pid}, {metadata.host}:{metadata.port})")
    print(f"PID file: {args.pid_file.expanduser()}")
    print(f"Log file: {Path(metadata.log_file).expanduser()}")


def _run_shutdown(args: argparse.Namespace) -> None:
    metadata, stale = load_running_metadata(args.pid_file)
    if metadata is None:
        if stale:
            print(f"No running background server found. Removed stale PID file {args.pid_file.expanduser()}.")
        else:
            print("codex-lb-cinamon background server is not running.")
        raise SystemExit(1)

    try:
        stopped = shutdown_background_server(args.pid_file, timeout_seconds=args.timeout)
    except RuntimeError as exc:
        raise SystemExit(str(exc)) from exc
    if stopped is None:
        print("codex-lb-cinamon background server is not running.")
        raise SystemExit(1)

    print(f"Stopped codex-lb-cinamon background server (pid {stopped.pid}).")


def main() -> None:
    args = _parse_args()
    command = args.command

    if command == "serve":
        _run_foreground(_serve_options_from_args(args))
        return
    if command == "start":
        _run_background_start(args)
        return
    if command == "status":
        _run_status(args)
        return
    if command == "shutdown":
        _run_shutdown(args)
        return

    raise SystemExit(f"Unsupported command: {command}")


if __name__ == "__main__":
    main()



from __future__ import annotations

from sqlalchemy.engine import make_url


def to_sync_database_url(database_url: str) -> str:
    parsed = make_url(database_url)
    driver = parsed.drivername

    if driver == "sqlite+aiosqlite":
        parsed = parsed.set(drivername="sqlite")
    elif driver == "postgresql+asyncpg":
        parsed = parsed.set(drivername="postgresql+psycopg")

    return parsed.render_as_string(hide_password=False)
